"""
Image IO

"""
import glob
import json
import os
import pathlib
import re

import ants
import nibabel
import numpy as np
import pandas as pd
from nibabel.filebasedimages import FileBasedHeader

from .constants import HALF_LIVES


def write_dict_to_json(meta_data_dict: dict, out_path: str):
    """
    Save a metadata file in python to a directory.

    Args:
        meta_data_dict (dict): A dictionary with imaging metadata, to be saved to file.
        out_path (str): Directory to which `meta_file` is to be saved.
    """
    with open(out_path, 'w', encoding='utf-8') as copy_file:
        json.dump(meta_data_dict, copy_file, indent=4)


def gen_meta_data_filepath_for_nifti(nifty_path:str):
    """
    Generates the corresponding metadata file path for a given nifti file path.

    This function takes a nifti file path (with `.nii` or `.nii.gz` extension)
    and replaces the extension with `.json` to derive the expected metadata file path.

    Args:
        nifty_path (str): Path to the nifti file (with `.nii` or `.nii.gz` extension).

    Returns:
        str: The generated metadata file path with a `.json` extension.
    """
    meta_data_path = re.sub(r'\.nii\.gz$|\.nii$', '.json', nifty_path)
    return meta_data_path


def safe_load_meta(input_metadata_file: str) -> dict:
    """
    Function to load a generic metadata json file.

    Args:
        input_metadata_file (str): Metadata file to be read.

    Returns:
        metadata (dict): The metadata in dictionary format.
    """
    if not os.path.exists(input_metadata_file):
        raise FileNotFoundError(f"Metadata file {input_metadata_file} not found. Does it have a "
                                "different path?")

    with open(input_metadata_file, 'r', encoding='utf-8') as meta_file:
        metadata = json.load(meta_file)
    return metadata


def load_metadata_for_nifti_with_same_filename(image_path) -> dict:
    """
    Static method to load metadata. Assume same path as input image path.

    Args:
        image_path (str): Path to image for which a .json file of the
            same name as the file but with different extension exists.

    Returns:
        metadata (dict): Dictionary where keys are fields in the image
            metadata file and values correspond to values in those fields.

    Raises:
        FileNotFoundError: If the provided image path cannot be found in the directory.
        Additionally, occurs if the metadata .json file cannot be found.
    """
    if not os.path.exists(image_path):
        raise FileNotFoundError(f"Image file {image_path} not found.")

    meta_path = gen_meta_data_filepath_for_nifti(image_path)
    metadata = safe_load_meta(input_metadata_file=meta_path)

    return metadata


def flatten_metadata(metadata: dict) -> dict:
    """
    Given a metadata dictionary, return an identical dictionary with any list-like or dict-like
    data replaced with individual values. Useful when converting several JSON files into a TSV
    file.

    Args:
        metadata (dict): The metadata file that may contain lists of data.

    Returns:
        metadata_for_tsv (dict): The same metadata with list-like data replaced with individual
            values.

    Note:
        List-like data is replaced by renaming the key it appears in with ordinal values. E.g. if
        metadata contains a key named ``FitPars`` with value [4,6] then the function would create
        two new keys, FitPars_1 and Fit_Pars2 with values 4 and 6 respectively. Likewise, nested
        dictionaries are replaced by combining the two keys identifying the data with underscores.
        Function is not robust for doubly nested lists and dictionaries.
    """
    metadata_for_tsv = {}
    for key in metadata:
        data = metadata[key]
        if isinstance(data,list):
            for i,val in enumerate(data):
                key_new = f'{key}_{i+1}'
                metadata_for_tsv[key_new] = val
        elif isinstance(data,dict):
            for inner_key in data:
                key_new = f'{key}_{inner_key}'
                metadata_for_tsv[key_new] = data[inner_key]
        else:
            metadata_for_tsv[key] = metadata[key]
    return metadata_for_tsv


def safe_copy_meta(input_image_path: str,
                   out_image_path: str):
    """
    Copy the metadata file from input image, to one with the same name as the
    output file. Intended to be used in functions operating on images in order
    to ensure a metadata file is associated with each new image.
    
    Args:
        input_image_path (str): Path to the input file for the function
            generating a new image.
        out_image_path (str): Path to the output file written by the function.
    """
    copy_meta_path = gen_meta_data_filepath_for_nifti(out_image_path)
    meta_data_dict = load_metadata_for_nifti_with_same_filename(input_image_path)
    write_dict_to_json(meta_data_dict=meta_data_dict, out_path=copy_meta_path)

def get_half_life_from_radionuclide(meta_data_file_path: str) -> float:
    """
    Extracts the radionuclide half-life in seconds from a nifti metadata file. This function
    grabs the tracer radionuclide from the metadata and assumes a fixed half-life based on this.
    Code borrowed from:
    https://github.com/bilgelm/dynamicpet/blob/main/src/dynamicpet/petbids/petbidsjson.py.

    Args:
        meta_data_file_path (str): Path to the nifti metadata file.

    Returns:
        float: The radionuclide half-life extracted from the tracer radionuclide.

    Raises:
        FileNotFoundError: If the metadata file does not exist at the provided path.
        KeyError: If the 'TracerRadionuclide' key is not found in the metadata file.
    """
    meta_data = safe_load_meta(meta_data_file_path)

    try:
        radionuclide = meta_data['TracerRadionuclide'].lower().replace("-", "")
    except KeyError as exc:
        raise KeyError("Required BIDS metadata field 'TracerRadionuclide' not found.") from exc

    return float(HALF_LIVES[radionuclide])


def get_half_life_from_meta(meta_data_file_path: str):
    """
    Extracts the radionuclide half-life (usually in seconds) from a nifti metadata file.

    Args:
        meta_data_file_path (str): Path to the nifti metadata file.

    Returns:
        float: The radionuclide half-life extracted from the metadata file.

    Raises:
        FileNotFoundError: If the metadata file does not exist at the provided path.
        KeyError: If the 'RadionuclideHalfLife' key is not found in the metadata file.
    """
    meta_data = safe_load_meta(meta_data_file_path)

    try:
        half_life = meta_data['RadionuclideHalfLife']
        return float(half_life)
    except KeyError as exc:
        raise KeyError("RadionuclideHalfLife not found in meta-data file.") from exc


def get_half_life_from_nifti(image_path:str):
    """
    Retrieves the radionuclide half-life from a nifti image file.

    This function first checks if the provided nifti image file exists. It then derives
    the corresponding metadata file path using :func:`_gen_meta_data_filepath_for_nifti`
    and finally retrieves the half-life from the metadata using :func:`get_half_life_from_meta`.

    Args:
        image_path (str): Path to the nifti image file.

    Returns:
        float: The radionuclide half-life extracted from the metadata file.

    Raises:
        FileNotFoundError: If the nifti image file does not exist at the provided path.
    """
    if not os.path.exists(image_path):
        raise FileNotFoundError(f"Image file {image_path} not found")
    meta_path = gen_meta_data_filepath_for_nifti(image_path)
    try:
        half_life = get_half_life_from_radionuclide(meta_path)
    except KeyError:
        half_life = get_half_life_from_meta(meta_path)
    return float(half_life)



class ImageIO:
    """
    :class:`ImageIO` to handle reading and writing imaging data and metadata.

    Provides several tools designed for reading and writing data within the Python environment.

    Key methods include:
        - :meth:`save_nii`: Saves a loaded NIfTI file to a file path.
        - :meth:`extract_image_from_nii_as_numpy`: Extracts imaging data from a NIfTI file as a numpy array.
        - :meth:`extract_header_from_nii`: Extracts header information from a NIfTI file as a dictionary.
        - :meth:`extract_np_to_nibabel`: Wraps imaging information in numpy into an Nibabel image.

    Attributes:
        verbose (bool): Set to `True` to output processing information.
    """

    def __init__(self, verbose: bool = True, ):
        """
        Initializes :class:`ImageIO` and sets verbose.

        Args:
            verbose (bool): Set to True to print debugging info to shell. Defaults to True.
        """
        self.verbose = verbose

    def save_nii(self, image: nibabel.nifti1.Nifti1Image, out_file: str):
        """
        Wrapper to save nifti to file.

        Args:
            image (nibabel.nifti1.Nifti1Image): Nibabel-type image to write to file.
            out_file (str): File path to which image will be written.
        """
        nibabel.save(image, out_file)
        if self.verbose:
            print(f"(ImageIO): Image saved to {out_file}")

    def extract_image_from_nii_as_numpy(self, image: nibabel.nifti1.Nifti1Image) -> np.ndarray:
        """
        Convenient wrapper to extract data from a .nii or .nii.gz file as a numpy array.

        Args:
            image (nibabel.nifti1.Nifti1Image): Nibabel-type image to write to file.

        Returns:
            The data contained in the .nii or .nii.gz file as a numpy array.
        """
        image_data = image.get_fdata()

        if self.verbose:
            print(f"(ImageIO): Image has shape {image_data.shape}")

        return image_data

    def extract_header_from_nii(self, image: nibabel.nifti1.Nifti1Image) -> FileBasedHeader:
        """
        Convenient wrapper to extract header information from a .nii or .nii.gz
        file as a nibabel file-based header.

        Args:
            image (nibabel.nifti1.Nifti1Image): Nibabel-type image to write to file.

        Returns:
            image_header (FileBasedHeader): The nifti header.
        """
        image_header = image.header

        if self.verbose:
            print(f"(ImageIO): Image header is: {image_header}")

        return image_header

    def extract_np_to_nibabel(self,
                              image_array: np.ndarray,
                              header: FileBasedHeader,
                              affine: np.ndarray) -> nibabel.nifti1.Nifti1Image:
        """
        Wrapper to convert an image array into nibabel object.

        Args:
            image_array (np.ndarray): Array containing image data.
            header (FileBasedHeader): Header information to include.
            affine (np.ndarray): Affine information we need to keep when rewriting image.

        Returns:
            image_nibabel (nibabel.nifti1.Nifti1Image): Image stored in nifti-like nibabel format.
        """
        image_nibabel = nibabel.nifti1.Nifti1Image(image_array, affine, header)
        return image_nibabel

    @staticmethod
    def affine_parse(image_affine: np.ndarray) -> tuple:
        """
        Parse the components of an image affine to return origin, spacing, direction.

        Args:
            image_affine (np.ndarray): A 4x4 affine matrix defining spacing, origin,
                and direction of an image.
        """
        spacing = nibabel.affines.voxel_sizes(image_affine)
        origin = image_affine[:, 3]

        quat = nibabel.quaternions.mat2quat(image_affine[:3, :3])
        dir_3x3 = nibabel.quaternions.quat2mat(quat)
        direction = np.zeros((4, 4))
        direction[-1, -1] = 1
        direction[:3, :3] = dir_3x3

        return spacing, origin, direction

    def extract_np_to_ants(self, image_array: np.ndarray, affine: np.ndarray) -> ants.ANTsImage:
        """
        Wrapper to convert an image array into ants object.
        Note header info is lost as ANTs does not carry this metadata.

        Args:
            image_array (np.ndarray): Array containing image data.
            affine (np.ndarray): Affine information we need to keep when rewriting image.

        Returns:
            image_ants (ants.ANTsImage): Image stored in nifti-like nibabel format.
        """
        origin, spacing, direction = self.affine_parse(affine)
        image_ants = ants.from_numpy(data=image_array, spacing=spacing, origin=origin, direction=direction)
        return image_ants


def read_label_map_tsv(label_map_file: str) -> dict | pd.DataFrame:
    """
    Static method to read a label map, translating region indices to region names,
    as a dictionary. Assumes tsv format.

    Args:
        label_map_file (str): Path to a json-formatted label map file.

    Returns:
        label_map (pd.DataFrame): Dataframe matching region indices, names,
            abbreviations, and mappings.

    Raises:
        FileNotFoundError: If the provided ctab file cannot be found in the directory.
    """
    if not os.path.exists(label_map_file):
        raise FileNotFoundError(f"Image file {label_map_file} not found")

    label_map = pd.read_csv(label_map_file, sep='\t')

    return label_map


def safe_load_4dpet_nifti(filename: str) -> nibabel.nifti1.Nifti1Image:
    """
    Safely load a 4D PET NIfTI file.

    This function checks if the given file has a '.nii' or '.nii.gz' extension, then tries to load
    it as a NIfTI file using the nibabel library. If the file cannot be loaded, it raises an
    exception.

    Args:
        filename (str): The path of the NIfTI file to be loaded.

    Returns:
        Nifti1Image: The loaded NIfTI 4D PET image.

    Raises:
        ValueError: If the file does not have a '.nii' or '.nii.gz' extension.
        Exception:  If an error occurred while loading the NIfTI file.
    """
    if not filename.endswith(('.nii', '.nii.gz')):
        raise ValueError(
            "Invalid file extension. Only '.nii' and '.nii.gz' are supported.")

    try:
        return nibabel.load(filename=filename)
    except Exception as e:
        print(f"Couldn't read file {filename}. Error: {e}")
        raise e


def validate_two_images_same_dimensions(image_1: nibabel.nifti1.Nifti1Image,
                                        image_2: nibabel.nifti1.Nifti1Image,
                                        check_4d: bool=False):
    """
    Check the dimensions of two Nifti1Image objects and verify they have the same shape.

    Args:
        image_1 (nibabel.nifti1.Nifti1Image): The first image of the two to check image size.
        image_2 (nibabel.nifti1.Nifti1Image): The second image of the two to check image size.
        check_4d (bool): If true, checks all dimensions including validating the number of frames.
            If false, only checks first three dimensions. Default False.
    
    Raises:
        ValueError: If images do not have the same dimensions.
    """
    shape_1 = image_1.shape
    shape_2 = image_2.shape

    same_shape = False
    if check_4d:
        same_shape = shape_1 == shape_2
    else:
        same_shape = shape_1[:3] == shape_2[:3]

    if not same_shape:
        raise ValueError(f'Got incompatible image sizes: {shape_1}, {shape_2}.')

def infer_sub_ses_from_tac_path(tac_path: str):
    """
    Infers subject and session IDs from a TAC file path by analyzing the filename.

    This method extracts subject and session IDs from the filename of a TAC file. It checks the
    presence of a `sub-` and `ses-` marker in the filename, which is followed by the subject and
    session respectively. This segment name is then formatted with each part capitalized. If no
    subject or session is found a generic value of `UNK` is returned.

    Args:
        tac_path (str): Path of the TAC file.
        tac_id (int): ID of the TAC.

    Returns:
        tuple: Inferred subject and session IDs.
    """
    path = pathlib.Path(tac_path)
    assert path.suffix == '.tsv', '`tac_path` must point to a TSV file (*.tsv)'
    filename = path.name
    fileparts = filename.split("_")
    subname = 'XXXX'
    for part in fileparts:
        if 'sub-' in part:
            subname = part.split('sub-')[-1]
            break
    if subname == 'XXXX':
        subname = 'UNK'
    else:
        name_parts = subname.split("-")
        subname = ''.join(name_parts)
    sesname = 'XXXX'
    for part in fileparts:
        if 'ses-' in part:
            sesname = part.split('ses-')[-1]
            break
    if sesname == 'XXXX':
        subname = 'UNK'
    else:
        name_parts = sesname.split("-")
        sesname = ''.join(name_parts)
    return subname, sesname


def km_regional_fits_to_tsv(fit_results_dir: str, out_tsv_dir: str):
    """
    Tidies the output of regional kinetic modeling results by converting JSON files into a TSV file
    with one row per fit region. Accomodates lists by converting them into key-value pairs. Assigns
    a subject and session to each row inferred from the original TAC file path.

    Requires fields to be identical across all JSON results files.

    Args:
        fit_results_dir (str): Directory where RTM results are stored in JSON files.
        out_tsv_dir (str): Path where resulting TSV file containing fit results will be stored.

    Returns:
        km_fits (pd.DataFrame): DataFrame containing KM fit data for all regions.
    """
    fit_results_jsons = glob.glob(os.path.join(fit_results_dir,'*.json'))
    km_fits = pd.DataFrame()
    for i,fit in enumerate(fit_results_jsons):
        fit_load = safe_load_meta(fit)
        fit_clean = flatten_metadata(fit_load)
        sub, ses = infer_sub_ses_from_tac_path(fit_clean['FilePathTTAC'])
        fit_clean['sub_id'] = sub
        fit_clean['ses_id'] = ses
        fit_pd = pd.DataFrame(fit_clean,index=[i])
        km_fits = pd.concat([km_fits,fit_pd])
    km_fits.to_csv(out_tsv_dir,sep='\t')
    return km_fits



# # Configuration file for the Sphinx documentation builder.
#
from datetime import datetime

project = 'PETPAL (Positron Emission Tomography Analysis Library)'
year = datetime.today().year
copyright = f'2024-{year}, Furqan Dar, Bradley Judge, Noah Goldman, Kenan Oestreich'
author = 'Furqan Dar, Bradley Judge, Noah Goldman, Kenan Oestreich'
release = '0.1.0'

extensions = [
    'autoapi.extension',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.ifconfig',
    'matplotlib.sphinxext.plot_directive',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_design',
]

language = 'en'

html_static_path = ['_static']
html_title = 'PETPAL'
html_split_index = False
html_theme = 'pydata_sphinx_theme'
html_theme_options = {
    "pygments_light_style": "default",
    "pygments_dark_style": "monokai",
    "secondary_sidebar_items": {
        "**/*": ["page-toc", "edit-this-page", "sourcelink"]
    },
    "show_toc_level" : 3,
    "icon_links": [
            {
            "name": "GitHub",
            "url" : "https://github.com/FurqanDar/PETPAL",
            "icon": "fa-brands fa-github",
            },
                 ],
    "navbar_align": "left",
    "navbar_center": ["navbar-nav"],
}

exclude_patterns = ['_build']
templates_path = ["_templates"]
source_suffix = '.rst'
master_doc = 'index'

toc_object_entries_show_parents = "hide"
todo_include_todos = True

# autoapi configuration
autoapi_type = 'python'
autoapi_dirs = ['../petpal']

# Options: https://sphinx-autoapi.readthedocs.io/en/latest/reference/config.html#customisation-options
autoapi_options = [
    'members',
    'undoc-members',
    'inherited-members',
    'private-members',
    'special-members',
    'show-inheritance',
    'show-module-summary'
]

autoapi_ignore = ['*cli*']
autoapi_own_page_level = 'function'
autoapi_template_dir = '_templates'
autoapi_python_class_content = 'both'
autoapi_member_order = 'bysource'
autoapi_keep_files = True
autoapi_generate_api_docs = True
autoapi_toctree_entries = True
autoapi_use_implicit_namespaces= False

napoleon_use_ivar = True
napoleon_use_rtype = False

intersphinx_mapping = {
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'numba': ('https://numba.readthedocs.io/en/stable/', None),
    'sklearn': ('https://scikit-learn.org/stable/', None),
    'ants': ('https://antspy.readthedocs.io/en/stable/', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
    'lmfit': ('https://lmfit.github.io/lmfit-py/', None),
}

def skip_main_funcs(app, what, name, obj, skip, options):
    if "main" in name and what == "function":
       skip = True
    elif what=="attribute":
       skip = True
    return skip

def setup(sphinx):
   sphinx.connect("autoapi-skip-member", skip_main_funcs)


"""
This module provides a Command-line interface (CLI) for preprocessing imaging data to
produce regional PET Time-Activity Curves (TACs) and prepare data for parametric imaging analysis.

The user must provide:
    * The sub-command. Options: 'weighted-sum', 'motion-correct', 'register-pet', or 'write-tacs'.
    * Path to PET input data in NIfTI format. This can be source data, or with some preprocessing
      such as registration or motion correction, depending on the chosen operation.
    * Directory to which the output is written.
    * The name of the subject being processed, for the purpose of naming output files.
    * 3D imaging data, such as anatomical, segmentation, or PET sum, depending on the desired
      preprocessing operation.
    * Additional information needed for preprocessing, such as color table or half-life.
    

Examples:

   * Auto Crop:

   .. code-block:: bash

       petpal-preproc auto-crop -i /path/to/input_img.nii.gz -o petpal_crop.nii.gz -t 0.05


   * Windowed Motion Correction:

   .. code-block:: bash

       petpal-preproc motion-correct -i /path/to/input_img.nii.gz -o petpal_moco.nii.gz --window-size 120 --transform-type QuickRigid


   * Register to anatomical:

   .. code-block:: bash

       petpal-preproc register-pet -i /path/to/input_img.nii.gz -o petpal_reg.nii.gz --motion-target 0 600 --anatomical /path/to/anat.nii.gz


    * Write regional tacs:

   .. code-block:: bash

    petpal-preproc write-tacs -i /path/to/input_img.nii.gz -p sub-001 -o /tmp/petpal_tacs -s /path/to/segmentation.nii.gz -l perlcyno -x


   * Write regional tacs, deprecated:

   .. code-block:: bash

       petpal-preproc write-tacs -i /path/to/input_img.nii.gz -o /tmp/petpal_tacs --segmentation /path/to/segmentation.nii.gz --label-map-path /path/to/dseg.tsv


   * Half life weighted sum of series:

   .. code-block:: bash

       petpal-preproc weighted-series-sum -i /path/to/input_img.nii.gz -o petpal_wss.nii.gz --start-time 1800 --end-time 7200


   * SUVR Image:

   .. code-block:: bash

       petpal-preproc suvr -i /path/to/input_img.nii.gz -o petpal_suvr.nii.gz --segmentation /path/to/segmentation.nii.gz --ref-region 1


   * Gaussian blur image:

   .. code-block:: bash

       petpal-preproc gauss-blur -i /path/to/input_img.nii.gz -o petpal_blur.nii.gz --blur-size-mm 8


   * Divide image by scale factor:

   .. code-block:: bash

       petpal-preproc rescale-image -i /path/to/input_img.nii.gz -o petpal_rescale.nii.gz --scale-factor 1000


   * Warp image to atlas:

   .. code-block:: bash

       petpal-preproc warp-pet-atlas -i /path/to/input_img.nii.gz -o petpal_reg-atlas.nii.gz --anatomical /path/to/anat.nii.gz --reference-atlas /path/to/atlas.nii.gz

    * Crop segmentation to PET FOV:

    .. code-block:: bash

        petpal-preproc seg-crop -i /path/to/input_img.nii.gz -o petpal_cropped_seg.nii.gz --segmentation /path/to/segmentation.nii.gz

See Also:
    * :mod:`~petpal.preproc.image_operations_4d` - module used for operations on 4D images.
    * :mod:`~petpal.preproc.motion_corr` - module for motion correction tools.
    * :mod:`~petpal.preproc.register` - module for MRI and atlas registration.

"""
import argparse
import ants


from ..preproc import (image_operations_4d,
                       motion_corr,
                       register,
                       regional_tac_extraction,
                       standard_uptake_value,
                       segmentation_tools,
                       brain_mask_pet)


_PREPROC_EXAMPLES_ = r"""
Examples:
  - Auto crop:
    petpal-preproc auto-crop -i /path/to/input_img.nii.gz -o petpal_crop.nii.gz -t 0.05
  - Windowed moco:
    petpal-preproc motion-correct -i /path/to/input_img.nii.gz -o petpal_moco.nii.gz --window-size 120 --transform-type QuickRigid
  - Register to anatomical:
    petpal-preproc register-pet -i /path/to/input_img.nii.gz -o petpal_reg.nii.gz --motion-target 0 600 --anatomical /path/to/anat.nii.gz
  - Write regional tacs:
    petpal-preproc write-tacs -i /path/to/input_img.nii.gz -p sub-001 -o /tmp/petpal_tacs -s /path/to/segmentation.nii.gz -l perlcyno -x
  - Write tacs, deprecated:
    petpal-preproc write-tacs-old -i /path/to/input_img.nii.gz -o /tmp/petpal_tacs --segmentation /path/to/segmentation.nii.gz --label-map-path /path/to/dseg.tsv
  - Half life weighted sum of series:
    petpal-preproc weighted-series-sum -i /path/to/input_img.nii.gz -o petpal_wss.nii.gz --start-time 1800 --end-time 7200
  - SUVR:
    petpal-preproc suvr -i /path/to/input_img.nii.gz -o petpal_suvr.nii.gz --segmentation /path/to/segmentation.nii.gz --ref-region 1
  - Gauss blur:
    petpal-preproc gauss-blur -i /path/to/input_img.nii.gz -o petpal_blur.nii.gz --blur-size-mm 8
  - Divide image by scale factor:s
    petpal-preproc rescale-image -i /path/to/input_img.nii.gz -o petpal_rescale.nii.gz --scale-factor 1000
  - Warp to atlas:
    petpal-preproc warp-pet-atlas -i /path/to/input_img.nii.gz -o petpal_reg-atlas.nii.gz --anatomical /path/to/anat.nii.gz --reference-atlas /path/to/atlas.nii.gz
  - SUV:
    petpal-preproc suv -i /path/to/input_img.nii.gz -o petpal_suv.nii.gz --weight 75 --dose 250 --start-time 1200 --end-time 3600
  - Crop segmentation image to PET FOV:
    petpal-preproc seg-crop -i /path/to/input_img.nii.gz -o petpal_cropped_seg.nii.gz --segmentation /path/to/segmentation.nii.gz
  - Add eroded white matter region to segmentation image:
    petpal-preproc eroded-wm -i /path/to/input_segmentation.nii.gz -o petpal_seg_with_eroded_wm.nii.gz
  - Apply brain mask to PET image:
    petpal-preproc brain-mask-pet -i /path/to/input_img.nii.gz -o petpal_brain_masked.nii.gz --atlas $FSLDIR/data/standard/MNI152_T1_1mm.nii.gz --mask $FSLDIR/data/standard/MNI152_T1_1mm_brain_mask.nii.gz
"""


def _add_common_args(parser: argparse.ArgumentParser) -> None:
    """Adds common arguments ('--input-img', '--out-img') to a provided ArgumentParser
    object.

    This function modifies the passed ArgumentParser object by adding two arguments commonly
    used for the command line scripts. It uses the add_argument method of the ArgumentParser class.
    After running this function, the parser will be able to accept and parse these additional
    arguments from the command line when run.
    
    .. note::
        This function modifies the passed `parser` object in-place and does not return anything.

    Args:
        parser (argparse.ArgumentParser): The argument parser object to which the arguments are
            added.

    Raises:
        argparse.ArgumentError: If a duplicate argument tries to be added.

    Side Effects:
        Modifies the ArgumentParser object by adding new arguments.

    Example:
        .. code-block:: python

            parser = argparse.ArgumentParser()
            _add_common_args(parser)
            args = parser.parse_args(['--pet', 'pet_file', '--out-dir', './', '--prefix', 'prefix'])
            print(args.pet)
            print(args.out_dir)
            print(args.prefix)

    """
    parser.add_argument('-o',
                        '--out-img',
                        default='petpal_preproc_output.nii.gz',
                        help='Output image filename')
    parser.add_argument('-i', '--input-img',required=True,help='Path to input image.',type=str)


def _generate_args() -> argparse.ArgumentParser:
    """
    Generates command line arguments for method :func:`main`.

    Returns:
        args (argparse.ArgumentParser): Arguments used in the command line and their corresponding
            values.
    """
    parser = argparse.ArgumentParser(prog='petpal-preproc',
                                     description="Command line interface for running PET "
                                                 "pre-processing steps.",
                                     epilog=_PREPROC_EXAMPLES_,
                                     formatter_class=argparse.RawTextHelpFormatter)

    subparsers = parser.add_subparsers(dest="command", help="Sub-command help.")

    parser_wss = subparsers.add_parser('weighted-series-sum',
                                       help='Half-life weighted sum of 4D PET series.')
    _add_common_args(parser_wss)
    parser_wss.add_argument('--start-time',
                            required=False,
                            help='Start time of sum in seconds.',
                            type=float,
                            default=0)
    parser_wss.add_argument('--end-time',
                            required=False,
                            help='End time of sum in seconds.',
                            type=float,
                            default=-1)

    parser_crop = subparsers.add_parser('auto-crop',
                                        help='Automatically crop 4D PET image using threshold.')
    _add_common_args(parser_crop)
    parser_crop.add_argument('-t','--thresh-val', required=True,default=0.01,
                            help='Fractional threshold to crop image projections.',type=float)

    parser_tac = subparsers.add_parser('write-tacs',
                                       help='Write ROI TACs from 4D PET using segmentation masks.')
    parser_tac.add_argument('-i', '--input-img',required=True,help='Path to input image.',type=str)
    parser_tac.add_argument('-p',
                            '--patid',
                            help='Name of participant, appended to beginning of TAC output files.',
                            required=True,
                            type=str)
    parser_tac.add_argument('-o',
                            '--out-tac-dir',
                            required=True,
                            help='Output TAC folder dir')
    parser_tac.add_argument('-s', '--segmentation', required=True,
                            help='Path to segmentation image in anatomical space.')
    parser_tac.add_argument('-l',
                            '--label-map',
                            required=True,
                            help='Label map for the seg image, either a preset option or path to a '
                                 'json file. E.g. freesurfer, freesurfer_merge_lr, perlcyno, '
                                 'perlcyno_merge_lr, /path/to/my_label_map.json.')
    parser_tac.add_argument('-x',
                            '--excel',
                            action='store_true',
                            required=False,
                            default=False,
                            help='Option to store results as a single file table instead of one '
                                 'TAC file per region.')

    parser_oldtac = subparsers.add_parser('write-tacs-old',
                                          help='DEPRECATED Write ROI TACs from 4D PET using '
                                               'segmentation masks. Uses `dseg.tsv`file as label '
                                               'map.')
    parser_oldtac.add_argument('-i',
                               '--input-img',
                               required=True,
                               help='Path to input image.',
                               type=str)
    parser_oldtac.add_argument('-o',
                            '--out-tac-dir',
                            default='petpal_tacs',
                            help='Output TAC folder dir')
    parser_oldtac.add_argument('-s', '--segmentation', required=True,
                            help='Path to segmentation image in anatomical space.')
    parser_oldtac.add_argument('-l',
                            '--label-map-path',
                            required=True,
                            help='Path to label map dseg.tsv')

    parser_warp = subparsers.add_parser('warp-pet-atlas',
                                        help='Perform nonlinear warp on PET to atlas.')
    _add_common_args(parser_warp)
    parser_warp.add_argument('-a',
                             '--anatomical',
                             required=True,
                             help='Path to 3D anatomical image (T1w or T2w).',
                            type=str)
    parser_warp.add_argument('-r',
                             '--reference-atlas',
                             required=True,
                             help='Path to anatomical atlas.',
                             type=str)

    parser_suvr = subparsers.add_parser('suvr',help='Compute SUVR on a parametric PET image.')
    _add_common_args(parser_suvr)
    parser_suvr.add_argument('-s', '--segmentation', required=True,
                            help='Path to segmentation image in anatomical space.')
    parser_suvr.add_argument('-r',
                             '--ref-region',
                             help='Reference region to normalize SUVR to. List multiple regions to '
                                  'combine as one reference.',
                             required=True,
                             type=int,
                             nargs='+')
    parser_suvr.add_argument('--start-time',
                            required=True,
                            help='Start time for SUVR calculation in seconds from scan start',
                            type=float)
    parser_suvr.add_argument('--end-time',
                            required=True,
                            help='End time for SUVR calculation in seconds from scan start',
                            type=float)

    parser_blur = subparsers.add_parser('gauss-blur',help='Perform 3D gaussian blurring.')
    _add_common_args(parser_blur)
    parser_blur.add_argument('-b',
                             '--blur-size-mm',
                             help='Size of gaussian kernal with which to blur image.',
                             type=float)

    parser_rescale = subparsers.add_parser('rescale-image',help='Divide an image by a scalar.')
    _add_common_args(parser_rescale)
    parser_rescale.add_argument('-r',
                                '--scale-factor',
                                help='Divides image by this number',
                                type=float,
                                required=True)


    parser_window_moco = subparsers.add_parser('motion-correct',
                                               help='Correct for inter-frame motion in dynamic PET '
                                                    'images')
    _add_common_args(parser_window_moco)
    parser_window_moco.add_argument('--motion-target', default=None, nargs='+',
                            help="Motion target option. Can be an image path, "
                                 "'weighted_series_sum' or a tuple "
                                 "(i.e. '--motion-target 0 600' for first ten minutes).",
                            required=True)
    parser_window_moco.add_argument('-w', '--window-size', default=60.0, type=float,
                                    help="Window size in seconds.",)
    xfm_types = ['QuickRigid', 'Rigid', 'DenseRigid', 'Affine', 'AffineFast']
    parser_window_moco.add_argument('-y', '--transform-type', default='QuickRigid', type=str,
                                    choices=xfm_types,
                                    help="Type of ANTs transformation to apply when registering.")

    parser_reg = subparsers.add_parser('register-pet',
                                       help='Register 4D PET to MRI anatomical space.')
    _add_common_args(parser_reg)
    parser_reg.add_argument('-a',
                            '--anatomical',
                            required=True,
                            help='Path to 3D anatomical image (T1w or T2w).',
                            type=str)
    parser_reg.add_argument('-t', '--motion-target', default=None, nargs='+',
                            help="Motion target option. Can be an image path, "
                                 "'weighted_series_sum' or a tuple (i.e. '-t 0 600' for first "
                                 "ten minutes).")

    parser_suv = subparsers.add_parser('suv',help='Standard Uptake Value (SUV) calculation')
    _add_common_args(parser_suv)
    parser_suv.add_argument('-w',
                            '--weight',
                            required=True,
                            help='Weight of the participant in kg',
                            type=float)
    parser_suv.add_argument('-d',
                            '--dose',
                            required=True,
                            help='Dose of radiotracer injected in MBq',
                            type=float)
    parser_suv.add_argument('--start-time',
                            required=True,
                            help='Start time for SUV calculation in seconds from scan start',
                            type=float)
    parser_suv.add_argument('--end-time',
                            required=True,
                            help='End time for SUV calculation in seconds from scan start',
                            type=float)

    parser_seg_crop = subparsers.add_parser('seg-crop',help='Crop segmentation image to PET FOV')
    _add_common_args(parser_seg_crop)
    parser_seg_crop.add_argument('-s',
                                 '--segmentation',
                                 required=True,
                                 help='Path to segmentation image',
                                 type=str)

    parser_eroded_wm = subparsers.add_parser('eroded-wm',
                                             help='Add eroded white matter region to segmentation '
                                                  'image')
    _add_common_args(parser_eroded_wm)

    parser_brain_mask = subparsers.add_parser('brain-mask-pet',
                                              help='Apply brain mask to dynamic PET image')
    _add_common_args(parser_brain_mask)
    parser_brain_mask.add_argument('--atlas',
                                   required=True,
                                   help='Path to atlas template image. Can be PET or anatomical '
                                         'based.',
                                   type=str)
    parser_brain_mask.add_argument('--mask',
                                   required=True,
                                   help='Path to brain mask in atlas space.',
                                   type=str)

    return parser


def main():
    """
    Preproc command line interface
    """
    preproc_parser = _generate_args()
    args = preproc_parser.parse_args()

    if args.command is None:
        preproc_parser.print_help()
        raise SystemExit('Exiting without command')

    if 'motion_target' in dir(args):
        if len(args.motion_target)==1:
            motion_target = args.motion_target[0]
        else:
            motion_target = args.motion_target
    else:
        motion_target = None

    command = str(args.command).replace('-','_')

    match command:
        case 'weighted_series_sum':
            standard_uptake_value.weighted_sum_for_suv(input_image_path=args.input_img,
                                                       output_image_path=args.out_img,
                                                       start_time=args.start_time,
                                                       end_time=args.end_time)
        case 'auto_crop':
            image_operations_4d.SimpleAutoImageCropper(input_image_path=args.input_img,
                                                    out_image_path=args.out_img,
                                                    thresh_val=args.thresh_val,
                                                    verbose=True)
        case 'register_pet':
            register.register_pet(input_reg_image_path=args.input_img,
                                out_image_path=args.out_img,
                                reference_image_path=args.anatomical,
                                motion_target_option=motion_target,
                                verbose=True)
        case 'write_tacs_old':
            regional_tac_extraction.write_tacs(input_image_path=args.input_img,
                                            out_tac_dir=args.out_tac_dir,
                                            segmentation_image_path=args.segmentation,
                                            label_map_path=args.label_map_path,
                                            verbose=True)
        case 'write_tacs':
            tac_obj = regional_tac_extraction.WriteRegionalTacs(input_image_path=args.input_img,
                                                    segmentation_path=args.segmentation,
                                                    label_map=args.label_map)
            tac_obj(out_tac_prefix=args.patid,
                    out_tac_dir=args.out_tac_dir,
                    one_tsv_per_region=not args.excel)
        case 'warp_pet_atlas':
            register.warp_pet_to_atlas(input_image_path=args.input_img,
                                       anat_image_path=args.anatomical,
                                       atlas_image_path=args.reference_atlas,
                                       out_image_path=args.out_img,
                                       verbose=True)
        case 'gauss_blur':
            image_operations_4d.gauss_blur(input_image_path=args.input_img,
                                        blur_size_mm=args.blur_size_mm,
                                        out_image_path=args.out_img,
                                        verbose=True,
                                        use_fwhm=True)
        case 'suvr':
            standard_uptake_value.suvr(input_image_path=args.input_img,
                                       output_image_path=args.out_img,
                                       segmentation_image_path=args.segmentation,
                                       ref_region=args.ref_region,
                                       start_time=args.start_time,
                                       end_time=args.end_time)
        case 'motion_correct':
            motion_corrector = motion_corr.MotionCorrect()
            motion_corrector(input_image_path=args.input_img,
                             output_image_path=args.out_img,
                             motion_target_option=motion_target,
                             window_duration=args.window_size)
        case 'rescale_image':
            input_img = ants.image_read(filename=args.input_img)
            out_img = image_operations_4d.rescale_image(input_image=input_img,
                                                        rescale_constant=args.scale_factor)
            ants.image_write(image=out_img, filename=args.out_img)
        case 'suv':
            standard_uptake_value.suv(input_image_path=args.input_img,
                                      output_image_path=args.out_img,
                                      start_time=args.start_time,
                                      end_time=args.end_time,
                                      weight=args.weight,
                                      dose=args.dose)
        case 'seg_crop':
            input_img = ants.image_read(filename=args.input_img)
            seg_img = ants.image_read(filename=args.segmentation)
            seg_cropped = segmentation_tools.seg_crop_to_pet_fov(pet_img=input_img,
                                                                 segmentation_img=seg_img)
            ants.image_write(seg_cropped,args.out_img)
        case 'eroded_wm':
            segmentation_tools.eroded_wm_segmentation(input_segmentation_path=args.input_img,
                                                      out_segmentation_path=args.out_img)
        case 'brain_mask_pet':
            brain_mask_pet.brain_mask_pet(input_image_path=args.input_img,
                                          out_image_path=args.out_img,
                                          atlas_image_path=args.atlas,
                                          atlas_mask_path=args.mask)

if __name__ == "__main__":
    main()



import numpy as np
import pytest
from petpal.utils.math_lib import weighted_sum_computation
from petpal.utils.scan_timing import calculate_frame_reference_time
from petpal.preproc.decay_correction import calculate_frame_decay_factor

def test_weighted_sum_computation_all_ones_simple():
    # simple case: all voxels = 1, decay_correction = 1, frame_start[0]=0
    pet_series = np.ones((2, 2, 2, 3), dtype=float)
    frame_duration = np.array([10.0, 20.0, 30.0])
    half_life = 100.0
    frame_start = np.array([0.0, 10.0, 30.0])
    decay_correction = np.ones(3)

    decay_constant = np.log(2.0) / half_life
    image_total_duration = frame_duration.sum()
    total_decay = decay_constant * image_total_duration
    total_decay /= 1.0 - np.exp(-decay_constant * image_total_duration)
    total_decay /= np.exp(-decay_constant * frame_start[0])

    expected = np.full((2, 2, 2), total_decay)
    out = weighted_sum_computation(pet_series, frame_duration, half_life, frame_start, decay_correction)
    np.testing.assert_allclose(out, expected)

def test_weighted_sum_computation_manual_computation():
    # randomized small example, validate against manual numpy computation
    rng = np.random.default_rng(0)
    pet_series = rng.random((3, 2, 1, 4)).astype(float)
    frame_duration = np.array([5.0, 5.0, 10.0, 300.0])
    half_life = 50.0
    frame_start = np.array([1.0, 6.0, 11.0, 21.0])
    frame_ref_time = calculate_frame_reference_time(frame_duration=frame_duration,
                                                    frame_start=frame_start,
                                                    half_life=half_life)
    decay_correction = calculate_frame_decay_factor(frame_reference_time=frame_ref_time,
                                                    half_life=half_life)

    # manual expected computation following function logic
    decay_constant = np.log(2.0) / half_life
    image_total_duration = frame_duration.sum()
    total_decay = decay_constant * image_total_duration
    total_decay /= 1.0 - np.exp(-decay_constant * image_total_duration)
    total_decay /= np.exp(-decay_constant * frame_start[0])

    # compute weighted sum: sum_t(pet[...,t] * frame_duration[t] / decay_correction[t])
    scaled = pet_series * frame_duration / decay_correction  # broadcasting over last axis
    pet_series_sum_scaled = scaled.sum(axis=3)
    expected = pet_series_sum_scaled * total_decay / image_total_duration

    out = weighted_sum_computation(pet_series, frame_duration, half_life, frame_start, decay_correction)
    np.testing.assert_allclose(out, expected)


"""Utilities for metadata handling, scrubbing, etc..."""

from shutil import copyfile
from copy import deepcopy
from itertools import accumulate

from .image_io import safe_load_meta, write_dict_to_json
from .constants import HALF_LIVES
from .scan_timing import calculate_frame_reference_time
from ..preproc.decay_correction import calculate_frame_decay_factor


class BidsMetadataMender:
    """Class for repairing and filling in the gaps of BIDS metadata based on existing fields.

    For most use cases, just initialize a mender object, passing a string path to the .json file,
    and a boolean for whether the image was decay_corrected. Then, simply calling the object 
    (i.e. 'mender()') will result in keys being added/updated where possible.

    The following checks are made: 

    - If 'FrameDuration' is present, fill 'FrameTimesStart', assuming the first entry to be 0.
    - If 'TracerRadionuclide' is present, add 'RadionuclideHalfLife'.
        - Note that 'RadionuclideHalfLife' is not listed in BIDS, but we find it useful to store.
    - If all the previous keys were added, use them to calculate 'FrameReferenceTime'.
        - 'FrameReferenceTime' is not BIDS-required but it is useful.
    - If all the previous keys were added, add 'DecayCorrectionFactor' and 'ImageDecayCorrected'.
        - If decay_correction was set to False, 'DecayCorrectionFactor' is a list of ones of 
          len(FrameDuration) and 'ImageDecayCorrected will be 'false', per BIDS.

    Attributes: 
        metadata (dict): dictionary containing all the existing BIDS metadata
        filepath (str): path to .json file containing original BIDS metadata
        decay_correction (bool): whether or not the image is decay-corrected
    """
    metadata: dict
    filepath: str
    decay_correction: bool

    def __init__(self, json_filepath: str, decay_correction: bool = False):
        """Initializes a new BidsMetadataMender object.
        
        Args: 
            json_filepath (str): path to .json file with original BIDS metadata.
            decay_correction (bool): Whether or not the image is decay-corrected. Defaults to False.
        """
        self.metadata = safe_load_meta(input_metadata_file=json_filepath)
        self.filepath = json_filepath
        self.decay_correction = decay_correction

    def __call__(self, output_filepath: str | None = None):
        self._add_missing_keys()
        self._to_file(output_filepath)

    def _add_missing_keys(self):
        """Repair/Fill missing keys where possible."""
        updated_keys = []
        if 'FrameDuration' in self.metadata:
            self._add_frame_times_start()
            updated_keys.append('FrameTimesStart')
        if 'TracerRadionuclide' in self.metadata:
            self._add_half_life()
            updated_keys.append('RadionuclideHalfLife')
        if {'RadionuclideHalfLife', 'FrameDuration', 'FrameTimesStart'}.issubset(self.metadata):
            self._add_frame_reference_times()
            updated_keys.append('FrameReferenceTime')
        if self.decay_correction and {'RadionuclideHalfLife', 'FrameReferenceTime'}.issubset(self.metadata):
            self._add_decay_factors()
            updated_keys += ['DecayCorrectionFactor', 'ImageDecayCorrected']
        else:
            self._add_empty_decay_factors()
            updated_keys += ['DecayCorrectionFactor', 'ImageDecayCorrected']
        print(f'The following keys were updated: {updated_keys}')

    def _add_half_life(self):
        """Add "RadionuclideHalfLife" key to metadata."""
        metadata = deepcopy(self.metadata)
        radionuclide = metadata['TracerRadionuclide'].lower().replace("-", "")
        half_life = float(HALF_LIVES[radionuclide])
        metadata['RadionuclideHalfLife'] = half_life
        self.metadata = metadata

    def _add_empty_decay_factors(self):
        """Adds a list of ones for decay factors and sets 'ImageDecayCorrected' to False."""
        metadata = deepcopy(self.metadata)
        frame_durations = metadata['FrameDuration']
        decay_factors = [1 for i in frame_durations]
        metadata['DecayCorrectionFactor'] = decay_factors
        metadata['ImageDecayCorrected'] = 'False'
        self.metadata = metadata

    def _add_decay_factors(self):
        """Computes decay factors and adds 'DecayCorrectionFactor' to metadata."""
        metadata = deepcopy(self.metadata)
        half_life = metadata['RadionuclideHalfLife']
        decay_factors = [calculate_frame_decay_factor(frame_reference_time=t, half_life=half_life) for t in metadata['FrameReferenceTime']]
        metadata['DecayCorrectionFactor'] = decay_factors
        metadata.pop('DecayFactor', None)
        metadata['ImageDecayCorrected'] = 'True'
        self.metadata = metadata

    def _add_frame_times_start(self):
        """Fill in frame starts from frame durations, assuming first frame starts at 0."""
        metadata = deepcopy(self.metadata)
        frame_durations = metadata['FrameDuration']
        frame_starts = [0]
        frame_starts = frame_starts + list(accumulate(frame_durations[:-1]))
        metadata['FrameTimesStart'] = frame_starts
        self.metadata = metadata

    def _add_frame_reference_times(self):
        """Fill in frame reference times from frame starts and durations."""
        metadata = deepcopy(self.metadata)
        half_life = metadata['RadionuclideHalfLife']
        frame_starts = metadata['FrameTimesStart']
        frame_durations = metadata['FrameDuration']
        frame_reference_times = [calculate_frame_reference_time(frame_duration=duration, frame_start=start, half_life=half_life) for start, duration in zip(frame_starts, frame_durations)]
        metadata['FrameReferenceTime'] = frame_reference_times
        self.metadata = metadata

    def _to_file(self, filepath: str | None = None):
        """Write metadata dictionary to a .json file; defaults to making a backup."""
        if filepath is None:
            filepath = self.filepath
            copyfile(src=filepath, dst=filepath+".bak")
        write_dict_to_json(meta_data_dict=self.metadata, out_path=filepath)



"""
Command-line interface (CLI) for Partial Volume Correction (PVC) using the Symmetric Geometric Transfer Matrix (sGTM)
method. It uses argparse to handle command-line arguments and chooses the appropriate method based on the provided input.
The user must provide:
    * PET image file path
    * Segmentation image file path
    * FWHM for Gaussian blurring
    * (Optional) Segmentation label map table path.
Example usage:
    Using SGTM method:
        .. code-block:: bash
            pet-cli-pvc --method SGTM --pet-path /path/to/pet_image.nii --roi-path /path/to/roi_image.nii --fwhm (8.0, 7.0, 7.0)
See Also:
    SGTM and PETPVC methods implementation modules.
    :mod:`SGTM <pet_cli.symmetric_geometric_transfer_matrix>` - module for performing Symmetric Geometric Transfer Matrix PVC.
"""
import argparse

from ..utils.bids_utils import parse_path_to_get_subject_and_session_id
from ..preproc.symmetric_geometric_transfer_matrix import Sgtm


def sgtm_cli_run(input_image_path: str,
                 segmentation_image_path: str,
                 fwhm: float | tuple[float, float, float],
                 output_path: str,
                 segmentation_label_map_path: str | None = None,):
    """
    Apply the SGTM method for Partial Volume Correction.
    """
    sgtm_obj = Sgtm(input_image_path=input_image_path,
                    segmentation_image_path=segmentation_image_path,
                    label_map_option=segmentation_label_map_path,
                    fwhm=fwhm)
    sub_id, ses_id = parse_path_to_get_subject_and_session_id(path=input_image_path)
    sgtm_obj(output_path=output_path, out_tac_prefix=f'sub-{sub_id}_ses-{ses_id}')

def main():
    """
    Main function to handle command-line arguments and apply the appropriate PVC method.
    """
    parser = argparse.ArgumentParser(prog="PVC CLI",
                                     description="Apply Partial Volume Correction (PVC) to PET"
                                                 " images using sGTM. Works on 3D or 4D PET. 3D "
                                                 "result is the corrected uptake in each region, "
                                                 "4D result is the corrected TAC for each region.",
                                     epilog="Example of usage: pet-cli-pvc --pet-path"
                                            " /path/to/pet_image.nii --roi-path "
                                            "/path/to/roi_image.nii --fwhm 8.0")
    parser.add_argument("-i",
                        "--input-image",
                        required=True,
                        help="Path to the PET image file. Can be 3D or 4D.")
    parser.add_argument("-s",
                        "--segmentation_image",
                        required=True,
                        help="Path to the Segmentation image file.")
    parser.add_argument("-f",
                        "--fwhm",
                        required=True,
                        type=float,
                        help="Full Width at Half Maximum for Gaussian blurring (Tuple or single "
                             "float) in mm.")
    parser.add_argument("-t",
                        "--segmentation_label_map",
                        required=False, default=None,
                        help="Path to the table of segmentation labels map.")
    parser.add_argument("-o",
                        "--output",
                        required=True,
                        help="Path to PVC result. If input image is 3D, writes to a TSV file. If "
                             "input image is 4D, writes to a directory.")

    args = parser.parse_args()

    sgtm_cli_run(input_image_path=args.input_image,
                 segmentation_image_path=args.segmentation_image,
                 fwhm=args.fwhm,
                 output_path=args.output,
                 segmentation_label_map_path=args.segmentation_label_map)

if __name__ == "__main__":
    main()



"""
Provides functions for undo-ing decay correction and recalculating it.

"""

import math

import ants
import numpy as np

from ..utils import image_io
from ..utils.scan_timing import ScanTimingInfo


def undo_decay_correction(input_image_path: str,
                          output_image_path: str,
                          metadata_dict: dict = None) -> ants.ANTsImage:
    """Uses decay factors from the metadata for an image to remove decay correction for each frame.

    This function expects to find decay factors in the .json sidecar file, or the metadata_dict, if given. If there are
    no decay factors (either under the key 'DecayFactor' or the BIDS-required 'DecayCorrectionFactor') listed, it may
    result in unexpected behavior. In addition to returning an ANTsImage containing the "decay uncorrected" data, the
    function writes an image to output_image_path, unless it is passed as 'None'.

    Args:
        input_image_path (str): Path to input (.nii.gz or .nii) image. A .json sidecar file should exist in the same
             directory as the input image.
        output_image_path (str): Path to output (.nii.gz or .nii) output image. If None, no image will be written.
        metadata_dict (dict, optional): Metadata dictionary to use instead of corresponding .json sidecar. If not
            specified (default behavior), function will try to use sidecar .json in the same directory as
            input_image_path.

    Returns:
        ants.ANTsImage: ANTsImage with decay correction reversed."""

    decay_corrected_image = ants.image_read(filename=input_image_path)

    if metadata_dict is not None:
        json_data = metadata_dict
    else:
        json_data = image_io.load_metadata_for_nifti_with_same_filename(image_path=input_image_path)

    frame_info = ScanTimingInfo.from_nifti(image_path=input_image_path)
    decay_factors = frame_info.decay

    uncorrected_image_numpy = decay_corrected_image.numpy()

    for frame_num, decay_factor in enumerate(decay_factors):
        uncorrected_image_numpy[..., frame_num] /= decay_factor

    uncorrected_image = ants.from_numpy_like(data=uncorrected_image_numpy,
                                             image=decay_corrected_image)

    if output_image_path is not None:
        ants.image_write(image=uncorrected_image,
                         filename=output_image_path)

        json_data['DecayFactor'] = list(np.ones_like(decay_factors))
        json_data['ImageDecayCorrected'] = "false"
        output_json_path = image_io.gen_meta_data_filepath_for_nifti(nifty_path=output_image_path)
        image_io.write_dict_to_json(meta_data_dict=json_data,
                                    out_path=output_json_path)

    return uncorrected_image


def decay_correct(input_image_path: str,
                  output_image_path: str) -> ants.ANTsImage:
    r"""Recalculate decay_correction for nifti image based on frame reference times.

    This function will compute frame reference times based on frame time starts and frame durations (both of which
    are required by BIDS. These reference times are used in the following equation to determine the decay factor for
    each frame. For more information, refer to Turku Pet Centre's materials at
    https://www.turkupetcentre.net/petanalysis/decay.html

    .. math::
        decay\_factor = \exp(\lambda*t)

    where :math:`\lambda=\log(2)/T_{1/2}` is the decay constant of the radio isotope and depends on its half-life and
    `t` is the frame's reference time with respect to TimeZero (ideally, injection time).

    Note: BIDS 'DecayCorrectionTime' is set to 0 (seconds w.r.t. TimeZero) for the image. If this assumption doesn't
        hold, be wary of downstream effects.

    Args:
        input_image_path (str): Path to input (.nii.gz or .nii) image. A .json sidecar file should exist in the same
             directory as the input image.
        output_image_path (str): Path to output (.nii.gz or .nii) output image.

    Returns:
        ants.ANTsImage: Decay-Corrected Image

    """
    half_life = image_io.get_half_life_from_nifti(image_path=input_image_path)

    json_data = image_io.load_metadata_for_nifti_with_same_filename(image_path=input_image_path)
    uncorrected_image = ants.image_read(filename=input_image_path)

    frame_info = ScanTimingInfo.from_nifti(image_path=input_image_path)
    frame_reference_times = np.asarray(frame_info.start + frame_info.duration / 2.0, float).tolist()

    original_decay_factors = frame_info.decay
    if np.any(original_decay_factors != 1):
        raise ValueError(f'Decay Factors other than 1 found in metadata for {input_image_path}. This likely means the '
                         f'image has not had its previous decay correction undone. Try running undo_decay_correction '
                         f'before running this function to avoid decay correcting an image more than once.')

    corrected_data = uncorrected_image.numpy()
    new_decay_factors = []
    for frame_num, frame_reference_time in enumerate(frame_reference_times):
        new_decay_factor = math.exp(((math.log(2) / half_life) * frame_reference_time))
        corrected_data[..., frame_num] *= new_decay_factor
        new_decay_factors.append(new_decay_factor)

    corrected_image = ants.from_numpy_like(data=corrected_data,
                                           image=uncorrected_image)

    if output_image_path is not None:
        ants.image_write(image=corrected_image,
                         filename=output_image_path)
        output_json_path = image_io.gen_meta_data_filepath_for_nifti(nifty_path=output_image_path)
        json_data['DecayFactor'] = new_decay_factors
        json_data['ImageDecayCorrected'] = "true"
        json_data['ImageDecayCorrectionTime'] = 0
        json_data['FrameReferenceTime'] = frame_reference_times
        image_io.write_dict_to_json(meta_data_dict=json_data,
                                    out_path=output_json_path)

    return corrected_image


def calculate_frame_decay_factor(frame_reference_time: np.ndarray,
                                 half_life: float) -> np.ndarray:
    """Calculate decay correction factors for a scan given the frame reference time and half life.
    
    Important: 
        The frame reference time should be the time at which average activity occurs,
        not simply the midpoint. See
        :meth:`~petpal.utils.scan_timing.calculate_frame_reference_time` for more info.

    Args: 
        frame_reference_time (np.ndarray): Time at which the average activity occurs for the frame.
        half_life (float): Radionuclide half life.

    Returns: 
        np.ndarray: Decay Correction Factors for each frame in the scan.
    """
    decay_constant = np.log(2)/half_life
    frame_decay_factor = np.exp((decay_constant)*frame_reference_time)
    return frame_decay_factor



"""Module with function to get a motion target for motion correction and registration"""
import os
import tempfile
import ants

from ..utils.useful_functions import get_average_of_timeseries
from .standard_uptake_value import weighted_sum_for_suv


def determine_motion_target(motion_target_option: str | tuple | list,
                            input_image_path: str = None) -> str:
    """
    Produce a motion target given the ``motion_target_option`` from a method
    running registrations on PET, i.e. :meth:`motion_correction` or
    :meth:`register_pet`.

    The motion target option can be a string or a tuple. If it is a string,
    then if this string is a file, use the file as the motion target.

    If it is the option ``weighted_series_sum``, then run
    :meth:`weighted_series_sum` and return the output path.

    If it is the option ``mean_image``, then compute the time-average of the
    4D-PET image.

    If it is a tuple, run a weighted sum on the PET series on a range of
    frames. The elements of the tuple are treated as times in seconds, counted
    from the time of the first frame, i.e. (0,300) would average all frames
    from the first to the frame 300 seconds later. If the two elements are the
    same, returns the one frame closest to the entered time.

    Args:
        motion_target_option (str | tuple | list): Determines how the method behaves,
            according to the above description. Can be a file, a method
            ('weighted_series_sum' or 'mean_image'), or a tuple range e.g. (0,600).
        input_image_path (str): Path to the PET image. This is intended to
            be supplied by the parent method employing this function. Default
            value None.

    Returns:
        out_image_file (str): File to use as a target to compute
            transformations on.

    Raises:
        ValueError: If ``motion_target_option`` is not a string, list, or tuple. If it is a string,
            but does not match one of the preset options or path to a file, the error will also be
            raised.
        TypeError: If start and end time are incompatible with ``float`` type.
    """
    if isinstance(motion_target_option, str):
        if os.path.exists(motion_target_option):
            return motion_target_option

        if motion_target_option == 'weighted_series_sum':
            out_image_file = tempfile.mkstemp(suffix='_wss.nii.gz')[1]
            weighted_sum_for_suv(input_image_path=input_image_path,
                                output_image_path=out_image_file)
            return out_image_file

        if motion_target_option == 'mean_image':
            out_image_file = tempfile.mkstemp(suffix='_mean.nii.gz')[1]
            input_img = ants.image_read(input_image_path)
            mean_img = get_average_of_timeseries(input_image=input_img)
            ants.image_write(image=mean_img,filename=out_image_file)
            return out_image_file

        raise ValueError("motion_target_option did not match a file or 'weighted_series_sum'")

    if isinstance(motion_target_option, (list, tuple)):

        start_time = motion_target_option[0]
        end_time = motion_target_option[1]

        try:
            float(start_time)
            float(end_time)
        except Exception as exc:
            raise TypeError('Start time and end time of calculation must be '
                            'able to be cast into float! Provided values are '
                            f"{start_time} and {end_time}.") from exc

        out_image_file = tempfile.mkstemp(suffix='_wss.nii.gz')[1]
        weighted_sum_for_suv(input_image_path=input_image_path,
                                output_image_path=out_image_file,
                                start_time=float(start_time),
                                end_time=float(end_time))

        return out_image_file

    raise ValueError('motion_target_option did not match str or tuple type.')



"""Tools for preparing PET data for kinetic modeling and visualization"""
from . import motion_target
from . import image_operations_4d
from . import motion_corr
from . import partial_volume_corrections
from . import register
from . import standard_uptake_value
from . import symmetric_geometric_transfer_matrix
from . import segmentation_tools
from . import decay_correction
from . import regional_tac_extraction
from . import brain_mask_pet

def main():
    print("PETPAL - Pre-processing")


if __name__ == "__main__":
    main()



"""
Command-line interface (CLI) for generating graphical analysis plots of PET Time-Activity Curves (TACs).

This module provides a CLI to work with the graphical_plots module. It uses argparse to handle command-line arguments.

The user must provide:
    * Input TAC file path
    * Region of Interest (ROI) TAC file path
    * Threshold in minutes (below which data points will be not be fit)
    * The method name for generating the plot. Supported methods are 'patlak', 'logan', or 'alt-logan'.
    * Output directory where the plots will be saved

An optional filename prefix for the output files can also be supplied.

This script uses the 'Plot' class from the 'graphical_plots' module to generate and save the plots.

Example:
    .. code-block:: bash
    
        petpal-graph-plot --input-tac-path /path/to/input.tac --roi-tac-path /path/to/roi.tac --threshold-in-mins 30.0 --method-name patlak --output-directory ./plots --output-filename-prefix plot

See Also:
    :mod:`petpal.visualization.graphical_plots` - module for creating and saving graphical analysis plots of PET TACs.
    
"""

import argparse
from ..visualizations import graphical_plots as pet_plt


def main():
    parser = argparse.ArgumentParser(prog="Graphical Analysis Plots",
                                     description="Generate graphical analysis plots of PET TACs.",
                                     epilog="Example: petpal-graph-plot "
                                            "--input-tac-path /path/to/input.tac --roi-tac-path "
                                            "/path/to/roi.tac --threshold-in-mins 30.0 --method-name patlak "
                                            "--output-directory /path/to/output --output-filename-prefix plot")
    
    grp_io = parser.add_argument_group("I/O Paths And Prefixes")
    grp_io.add_argument("-i", "--input-tac-path", required=True, help="Path to the input TAC file.")
    grp_io.add_argument("-r", "--roi-tac-path", required=True, help="Path to the Region of Interest (ROI) TAC file.")
    grp_io.add_argument("-o", "--output-directory", required=True,
                        help="Path to the directory where the plot output will be saved.")
    grp_io.add_argument("-p", "--output-filename-prefix", default="",
                        help="An optional prefix for the output filenames.")
    
    grp_params = parser.add_argument_group("Method Parameters")
    grp_params.add_argument("-t", "--threshold-in-mins", required=True, type=float,
                        help="Threshold in minutes below which data points will be discarded.")
    grp_params.add_argument("-m", "--method-name", required=True, choices=['patlak', 'logan', 'alt-logan'],
                        help="Name of the method for generating the plot.")
    
    args = parser.parse_args()
    
    grph_plot = pet_plt.Plot(input_tac_path=args.input_tac_path, roi_tac_path=args.roi_tac_path,
                             threshold_in_mins=args.threshold_in_mins, method_name=args.method_name,
                             output_directory=args.output_directory, output_filename_prefix=args.output_filename_prefix)
    
    grph_plot.save_figure()


if __name__ == "__main__":
    main()


"""
Methods applying to segmentations.

Available methods:
* :meth:`combine_regions_as_mask`: Merge regions in a segmentation image into a mask with value 1
* :meth:`resample_segmentation`: Resample a segmentation image to the affine of a 4D PET image.
* :meth:`vat_wm_ref_region`: Compute the white matter reference region for the VAT radiotracer.

"""
from numbers import Integral
import numpy as np
import ants
import nibabel
from nibabel import processing
import pandas as pd

from ..utils.dimension import (check_physical_space_for_ants_image_pair,
                               gen_nd_image_based_on_image_list)

from ..utils.useful_functions import get_average_of_timeseries
from ..utils import math_lib


def combine_regions_as_mask(segmentation_img: ants.core.ANTsImage | np.ndarray,
                            label: int | list[int]) -> ants.core.ANTsImage:
    """
    Create a mask from a segmentation image and one or more labels.

    If just one label is provided, this function will return a mask where values are 1 at voxels
    equal to that label, and 0 elsewhere. If the labels provided are in a list, the mask will be 1
    at voxels in the segmentation that are equal to any of the provided values, and zero elsewhere.

    Args:
        segmentation_img (ants.core.ANTsImage | np.ndarray): Image or array of brain regions.
        label (int | list[int]): Label or labels to mask the segmentation with.
    
    Returns:
        mask (ants.core.ANTsImage | np.ndarray): Image or array of mask on the provided labels.
            Output type matches the type used in ``segmentation_img``.

    Example:
            
        .. code-block:: python

            import ants
        
            from petpal.preproc.segmentation_tools import combine_regions_as_mask

            # Load the image
            seg_img = ants.image_read('/path/to/seg.nii.gz')

            # If the segmentation is FreeSurfer aparc+aseg, then region 12 is the Right Putamen
            right_putamen = combine_regions_as_mask(segmentation_img = seg_img, label=12)

            # If we want a mask of both the right and left putamen, use regions 12 and 51
            whole_putamen = combine_regions_as_mask(segmentation_img = seg_img, label=[12, 51])
    

    """
    if isinstance(label, Integral):
        label = [label]
    mask = sum(segmentation_img==l for l in label)
    return mask


def segmentations_merge(segmentation_primary: np.ndarray,
                        segmentation_secondary: np.ndarray,
                        regions: list) -> np.ndarray:
    """
    Merge segmentations by assigning regions to a primary segmentation image from a secondary
    segmentation. Region indices are pulled from the secondary into the primary from a list.

    Primary and secondary segmentations must have the same shape and orientation.

    Args:
        segmentation_primary (np.ndarray): The main segmentation to which new
            regions will be added.
        segmentation_secondary (np.ndarray): Distinct segmentation with regions
            to add to the primary.
        regions (list): List of regions to pull from the secondary to add to
            the primary.
    
    Returns:
        segmentation_primary (np.ndarray): The input segmentation with new
            regions added.
    """
    for region in regions:
        condition_above = segmentation_secondary > region - 0.1
        condition_below = segmentation_secondary < region + 0.1
        region_mask = condition_above & condition_below
        segmentation_primary[region_mask] = region
    return segmentation_primary


def binarize(input_image_numpy: np.ndarray,
             out_val: float=1):
    """
    Convert a segmentation image array into a mask by setting nonzero values
    to a uniform output value, typically one.

    Args:
        input_image_numpy (np.ndarray): Input image to be binarized to zero and
            another value.
        out_val (float): Uniform value output image is set to.
    
    Returns:
        bin_mask (np.ndarray): Image array of same shape as input, with values
            only zero and ``out_val``.
    """
    nonzero_voxels = (input_image_numpy > 1e-37) & (input_image_numpy < -1e-37)
    bin_mask = np.zeros(input_image_numpy.shape)
    bin_mask[nonzero_voxels] = out_val
    return bin_mask


def parcellate_right_left(segmentation_numpy: np.ndarray,
                          region: int,
                          new_right_region: int,
                          new_left_region: int) -> np.ndarray:
    """
    Divide a region within a segmentation image into right and left values.
    Assumes left and right sides are neatly subdivided by the image midplane,
    with right values below the mean value of the x-axis (zeroth axis) and left
    values above the mean value of the x-axis (zeroth axis).

    Intended to work with FreeSurfer segmentations on images loaded with
    nibabel. Use outside of these assumptions at your own risk.

    Args:
        segmentation_numpy (np.ndarray): Segmentation image array loaded with Nibabel, RAS+
            orientation
        region (int): Region index in segmentation image to be split into left and right.
        new_right_region (int): Region on the right side assigned to previous region.
        new_left_region (int): Region on the left side assined to previous region.

    Returns:
        split_segmentation (np.ndarray): Original segmentation image array with new left and right
            values.
    """
    seg_shape = segmentation_numpy.shape
    x_mid = (seg_shape[0] - 1) // 2

    seg_region = np.where(segmentation_numpy==region)
    right_region = seg_region[0] <= x_mid
    seg_region_right = tuple((seg_region[0][right_region],
                              seg_region[1][right_region],
                              seg_region[2][right_region]))

    left_region = seg_region[0] > x_mid
    seg_region_left = tuple((seg_region[0][left_region],
                             seg_region[1][left_region],
                             seg_region[2][left_region]))

    split_segmentation = segmentation_numpy
    split_segmentation[seg_region_right] = new_right_region
    split_segmentation[seg_region_left] = new_left_region

    return split_segmentation


def replace_probabilistic_region(segmentation_numpy: np.ndarray,
                                 segmentation_zooms: list,
                                 blur_size_mm: float,
                                 regions: list,
                                 regions_to_replace: list):
    """
    Runs a correction on a segmentation by replacing a list of regions with
    a set of nearby regions. This is accomplished by creating masks of the
    nearby regions, blurring them to create probabilistic segmentation maps,
    finding the highest probability nearby region in the region to replace,
    and replacing values with the respective nearby region.

    This is useful for protocols where there residual regions not intended to
    be carried forward after generating new regions or merging segmentations.

    Args:
        segmentation_numpy (np.ndarray): Input segmentation array.
        segmentation_zooms (list): X,Y,Z side length of voxels in mm.
        blur_size_mm (float): FWHM of Gaussian kernal used to blur regions.
        regions (list): List of region indices to replace residual regions.
        regions_to_replace (list): List of regions to be replaced by nearby
            regions listed in ``regions``.

    Returns:
        segmentation_numpy (np.ndarray): The input segmentation with replaced
            regions.
    """
    segmentations_combined = []
    for region in regions:
        region_mask = combine_regions_as_mask(segmentation_img=segmentation_numpy,
                                              label=[region])

        region_blur = math_lib.gauss_blur_computation(input_image=region_mask,
                                                      blur_size_mm=blur_size_mm,
                                                      input_zooms=segmentation_zooms,
                                                      use_fwhm=True)
        segmentations_combined += [region_blur]

    segmentations_combined_np = np.array(segmentations_combined)
    probability_map = np.argmax(segmentations_combined_np,axis=0)
    blend = combine_regions_as_mask(segmentation_img=segmentation_numpy,
                                    label=regions_to_replace)

    for i, region in enumerate(regions):
        region_match = (probability_map == i) & (blend > 0)
        segmentation_numpy[region_match] = region

    return segmentation_numpy


def resample_segmentation(input_image_path: str,
                          segmentation_image_path: str,
                          out_seg_path: str,
                          verbose: bool):
    """
    Resamples a segmentation image to the resolution of a 4D PET series image. Takes the affine 
    information stored in the PET image, and the shape of the image frame data, as well as the 
    segmentation image, and applies NiBabel's ``resample_from_to`` to resample the segmentation to
    the resolution of the PET image. This is used for extracting TACs from PET imaging where the 
    PET and ROI data are registered to the same space, but have different resolutions.

    Args:
        input_image_path (str): Path to a .nii or .nii.gz file containing a 4D
            PET image, registered to anatomical space, to which the segmentation file is resampled.
        segmentation_image_path (str): Path to a .nii or .nii.gz file containing a 3D segmentation
            image, where integer indices label specific regions.
        out_seg_path (str): Path to a .nii or .nii.gz file to which the resampled segmentation
            image is written.
        verbose (bool): Set to ``True`` to output processing information.
    """
    pet_image = nibabel.load(input_image_path)
    seg_image = nibabel.load(segmentation_image_path)
    pet_series = pet_image.get_fdata()
    image_first_frame = pet_series[:, :, :, 0]
    to_vox_map_tuple = (image_first_frame.shape, pet_image.affine)
    seg_resampled = processing.resample_from_to(from_img=seg_image,
                                                to_vox_map=to_vox_map_tuple,
                                                order=0)
    nibabel.save(seg_resampled, out_seg_path)
    if verbose:
        print(f'Resampled segmentation saved to {out_seg_path}')


def vat_wm_ref_region(input_segmentation_path: str,
                      out_segmentation_path: str | None) -> ants.ANTsImage:
    """
    Generates the cortical white matter reference region described in O'Donnell
    JL et al. (2024).
     
    Reference: O'Donnell JL et al. (2024). PET Quantification of [18F]VAT in Human Brain and Its 
    Test-Retest Reproducibility and Age Dependence. J Nucl Med. 2024 Jun 
    3;65(6):956-961. doi: 10.2967/jnumed.123.266860. PMID: 38604762; PMCID:
    PMC11149597.
    
    Requires FreeSurfer segmentation with original label mappings. 

    Args:
        input_segmentation_path (str): Path to segmentation on which white
            matter reference region is computed.
        out_segmentation_path (str): Path to which white matter reference
            region mask image is saved.

    Returns:
        wm_erode (ants.ANTsImage): Eroded white matter reference region mask image.
    """
    wm_regions = [2,41,251,252,253,254,255,77,3000,3001,3002,3003,3004,3005,
                  3006,3007,3008,3009,3010,3011,3012,3013,3014,3015,3016,3017,
                  3018,3019,3020,3021,3022,3023,3024,3025,3026,3027,3028,3029,
                  3030,3031,3032,3033,3034,3035,4000,4001,4002,4003,4004,4005,
                  4006,4007,4008,4009,4010,4011,4012,4013,4014,4015,4016,4017,
                  4018,4019,4020,4021,4022,4023,4024,4025,4026,4027,4028,4029,
                  4030,4031,4032,4033,4034,4035,5001,5002]
    csf_regions = [4,14,15,43,24]

    seg_img = ants.image_read(input_segmentation_path)

    wm_merged = combine_regions_as_mask(segmentation_img=seg_img,
                                        label=wm_regions)
    csf_merged = combine_regions_as_mask(segmentation_img=seg_img,
                                         label=csf_regions)
    wm_csf_merged = wm_merged + csf_merged

    wm_csf_blurred = ants.smooth_image(image=wm_csf_merged,
                                       sigma=9,
                                       sigma_in_physical_coordinates=True,
                                       FWHM=True,max_kernel_width=16)

    wm_csf_eroded = ants.threshold_image(image=wm_csf_blurred, low_thresh=0.95, binary=True)
    wm_erode = ants.mask_image(image=wm_merged, mask=wm_csf_eroded)

    if out_segmentation_path is not None:
        ants.image_write(image=wm_erode, filename=out_segmentation_path)

    return wm_erode


def eroded_wm_segmentation(input_segmentation_path: str,
                           out_segmentation_path: str | None,
                           eroded_wm_region_mapping: int = 1) -> ants.ANTsImage:
    """
    Generates eroded white matter region on a segmentation image and merges it into the image,
    saving result as a new segmentation image.

    Requires FreeSurfer segmentation with original label mappings.
    
    Args:
        input_segmentation_path (str): Path to input freesurfer segmentation, such as aparc+aseg or
            wmparc.
        out_segmentation_path (str): Path to output segmentation image with replaced values in
            eroded whited matter region.
        eroded_wm_region_mapping (int): Segmentation mapping for the eroded white matter region in
            the output image. Default "1".

    Returns:
        seg_img (ants.ANTsImage): Input segmentation image with values in eroded white matter
            replaced with `eroded_wm_region_mapping`.

    See also:
        :meth:`~petpal.preproc.segmentation_tools.vat_wm_ref_region` - function that generates the
            eroded white matter region.
    """
    wm_erode = vat_wm_ref_region(input_segmentation_path=input_segmentation_path,
                                 out_segmentation_path=None)
    seg_img = ants.image_read(input_segmentation_path)
    seg_img[wm_erode==1] = int(eroded_wm_region_mapping)
    if out_segmentation_path is not None:
        ants.image_write(image=seg_img, filename=out_segmentation_path)

    return seg_img


def vat_wm_region_merge(wmparc_segmentation_path: str,
                        out_image_path: str,
                        wm_ref_segmentation_path: str,
                        bs_segmentation_path: str = None):
    """
    Merge subcortical structures into a merged segmentation image according to
    the protocol for processing the VAT radiotracer.

    Args:
        wmparc_segmentation_path (str): Path to `wmparc` segmentation generated
            by FreeSurfer.
        out_image_path (str): Path to which output fused segmentation is saved.
        wm_ref_segmentation_path (str): Path to eroded white matter reference
            region generated by :meth:`vat_wm_ref_region`.
        bs_segmentation_path (str): Path to brainstem segmentation generated by
            FreeSurfer. If None, then skip.
    """
    wmparc = nibabel.load(wmparc_segmentation_path)
    wm_ref = nibabel.load(wm_ref_segmentation_path)

    wmparc_img = wmparc.get_fdata()
    wm_ref_img = wm_ref.get_fdata()

    zooms = wmparc.header.get_zooms()

    wmparc_split = parcellate_right_left(segmentation_numpy=wmparc_img,
                                         region=77,
                                         new_right_region=2,
                                         new_left_region=41)

    if bs_segmentation_path is None:
        wmparc_bs = wmparc_split
    else:
        bs = nibabel.load(bs_segmentation_path)
        bs_img = bs.get_fdata()
        wmparc_bs = segmentations_merge(segmentation_primary=wmparc_split,
                                        segmentation_secondary=bs_img,
                                        regions=[173,174,175])
    wmparc_bs_prob = replace_probabilistic_region(segmentation_numpy=wmparc_bs,
                                                  segmentation_zooms=zooms,
                                                  blur_size_mm=6,
                                                  regions=[257,15,165],
                                                  regions_to_replace=[16])

    wmparc_bs_wmref = segmentations_merge(segmentation_primary=wmparc_bs_prob,
                                          segmentation_secondary=wm_ref_img,
                                          regions=[1])
    if len(wmparc_bs_wmref.shape)==4:
        out_array = wmparc_bs_wmref[:,:,:,0]
    else:
        out_array = wmparc_bs_wmref
    out_file = nibabel.nifti1.Nifti1Image(dataobj=out_array,
                                          header=wmparc.header,
                                          affine=wmparc.affine)
    nibabel.save(out_file,out_image_path)


def gw_segmentation(freesurfer_path: str,
                    dseg_path: str,
                    output_path: str):
    """
    Creates a gray matter and a white matter mask based on FreeSurfer regions. Useful for PVC.
    """
    dseg = pd.read_csv(dseg_path,sep=r'\s+')
    freesurfer = ants.image_read(freesurfer_path)
    freesurfer_np = freesurfer.numpy()
    gm_map = np.zeros_like(freesurfer_np)
    wm_map = np.zeros_like(freesurfer_np)

    for i,mapping in enumerate(dseg['mapping']):
        gw_label = dseg['gray_white_matter'].iloc[i]
        region_seg = np.where(freesurfer_np==mapping)
        if gw_label == 0:
            gm_map[region_seg] = 1
        elif gw_label == 1:
            wm_map[region_seg] = 1

    gm_img = ants.from_numpy(data=gm_map,
                             origin=freesurfer.origin,
                             spacing=freesurfer.spacing,
                             direction=freesurfer.direction)
    wm_img = ants.from_numpy(data=wm_map,
                             origin=freesurfer.origin,
                             spacing=freesurfer.spacing,
                             direction=freesurfer.direction)
    gw_map_template = gen_nd_image_based_on_image_list([gm_img, wm_img])
    gw_map_4d = ants.list_to_ndimage(image=gw_map_template,image_list=[gm_img,wm_img])
    ants.image_write(gw_map_4d,output_path)


def subcortical_mask(input_seg_path: str,
                     output_seg_path: str=None,
                     subcortical_regions: list=None):
    """
    Gets a mask for subcortical regions from a FreeSurfer label image.

    Args:
        input_seg_path (str): Path to segmentation label image.
        output_seg_path (str): Path to which subcortical mask is saved.
        subcortical regions (list): Regions to include in the subcortical mask. Uses a built in
            list of subcortical mappings unless overridden by the user.

    Returns:
        subcortical_img (ants.ANTsImage): Subcortical mask image.

    .. important::
        * Subcortical mappings are assumed to correspond to a subset of FreeSurfer subcortical
            regions
        * Default regions: whole cerebellum, thalamus, caudate, putamen, pallidum, brainstem.
    """
    subcortical_mappings = [7,8,10,11,12,13,16,49,50,51,52,173,174,175]

    if subcortical_regions is None:
        subcortical_regions = subcortical_mappings

    segmentation = ants.image_read(input_seg_path)
    segmentation_np = segmentation.numpy()
    subcortical_mask_arr = np.zeros(segmentation_np.shape)
    for region in subcortical_regions:
        region_seg = np.where(segmentation_np==region)
        subcortical_mask_arr[region_seg] = 1
    subcortical_img = ants.from_numpy(
        data=subcortical_mask_arr,
        origin=segmentation.origin,
        spacing=segmentation.spacing,
        direction=segmentation.direction
    )

    if output_seg_path is not None:
        ants.image_write(subcortical_img,output_seg_path)

    return subcortical_img


def calc_vesselness_measure_image(input_image: ants.core.ANTsImage,
                                  sigma_min: float = 2.0,
                                  sigma_max: float = 8.0,
                                  alpha: float = 0.5,
                                  beta: float = 0.5,
                                  gamma: float = 5.0,
                                  morph_open_radius: int = 1,
                                  **hessian_func_kwargs) -> ants.core.ANTsImage:
    """
    Computes a vesselness measure image using Hessian-based objectness filtering.

    This function calculates the vesselness measure of a given 3D image using
    multi-scale Hessian filtering with specified parameters. We call the
    :func:`ants.hessian_objectness` after max-normalizing the input image.
    It enhances tubular structures like vessels, making them more pronounced
    in the output image. Optionally, a morphological opening operation can be
    applied to the result to refine the output and remove pepper-like artefacts.

    From the docs of :func:`ants.hessian_objectness`:
    '
    Based on the paper by Westin et al., "Geometrical
    Diffusion Measures for MRI from Tensor Basis Analysis" and Luca Antiga's
    Insight Journal paper http://hdl.handle.net/1926/576.
    '

    Args:
        input_image (ants.core.ANTsImage): Input 3D image for vesselness computation.
        sigma_min (float, optional): Minimum scale for multi-scale Hessian filtering
            (default: 2.0).
        sigma_max (float, optional): Maximum scale for multi-scale Hessian filtering
            (default: 8.0).
        alpha (float, optional): Alpha parameter for vesselness computation
            (default: 0.5).
        beta (float, optional): Beta parameter for vesselness computation
            (default: 0.5).
        gamma (float, optional): Gamma parameter for vesselness computation
            (default: 5.0).
        morph_open_radius (int, optional): Radius for the optional morphological
            opening operation (default: 1). If set to 0, no morphological opening
            will be applied.
        **hessian_func_kwargs: Additional keyword arguments for the Hessian
            objectness function.

    Returns:
        ants.core.ANTsImage: The vesselness-enhanced image.

    Notes:
        - Input image must be 3D; an assertion will fail if a non-3D image is provided.
        - The input image is normalized before processing to ensure robustness.
        - Morphological opening, if applied, uses a grayscale operation to refine
          the tubular structures.
        - The function has defaults for vesselness computation, but can be used to detect
          globular or plate-like structures as well.

    Raises:
        AssertionError: If the input image is not 3D.

    Workflow:
        1. Normalize the input image to have values between 0 and 1.
        2. Apply Hessian-based objectness filtering using the provided parameters
           (`sigma_min`, `sigma_max`, `alpha`, `beta`, `gamma`).
        3. Perform a morphological opening operation with the specified radius
           (`morph_open_radius`), if applicable.
        4. Return the computed vesselness image.
    """
    assert len(input_image.shape) == 3, "Input image must be 3D."

    tmp_img: ants.core.ANTsImage = input_image / input_image.max()
    hess_objectness_img = tmp_img.hessian_objectness(sigma_min=sigma_min,
                                                     sigma_max=sigma_max,
                                                     gamma=gamma,
                                                     alpha=alpha,
                                                     beta=beta,
                                                     **hessian_func_kwargs)
    if morph_open_radius > 0:
        hess_objectness_img = hess_objectness_img.morphology(operation='open',
                                                             radius=morph_open_radius,
                                                             mtype='grayscale')
    return hess_objectness_img


def calc_vesselness_mask_from_quantiled_vesselness(input_image: ants.core.ANTsImage,
                                                   min_quantile: float = 0.99,
                                                   morph_dil_radius: int = 0,
                                                   z_crop: int = 3) -> ants.core.ANTsImage:
    """
    Generates a binary vesselness mask from a given vesselness image using quantile-based
    thresholding.

    This function creates a binary mask by thresholding a vesselness image at a specified
    quantile of non-zero voxel values. Additionally, it allows for optional z-axis cropping
    and morphological dilation to refine the mask.

    Args:
        input_image (ants.core.ANTsImage): Input vesselness image.
        min_quantile (float, optional): Minimum quantile value for voxel thresholding
            (default: 0.99). Must be in the range [0, 1).
        morph_dil_radius (int, optional): Radius for morphological dilation to refine the
            mask (default: 0). No dilation is applied if set to 0.
        z_crop (int, optional): Number of slices to crop from the z-axis from the bottom
            (default: 3).

    Returns:
        ants.core.ANTsImage: Binary vesselness mask.

    Notes:
        - The input image must be an ANTs image containing vesselness measures.
        - The quantile value (`min_quartile`) determines the threshold value based on
          non-zero voxel intensities.
        - If `z_crop` is greater than 0, the z-axis is cropped from the top and bottom.
        - Morphological dilation is applied to the binary mask with the specified radius,
          if provided.
          
    Raises:
        - AssertionError: If the input image is not 3D.
        - AssertionError: If the provided quantile is not in the range [0, 1].
        - AssertionError: If the provided z-crop is larger than the number of z-slices in the
          input image.

    """
    assert 1 >= min_quantile >= 0, "Minimal quantile must be greater than 0 and less than 1."
    assert len(input_image.shape) == 3, "Input image must be 3D."
    assert z_crop < input_image.shape[2], "Z-crop must be less than input image's Z-dimension."

    vess_vals_arr = input_image.numpy()
    vess_vals_arr = vess_vals_arr[vess_vals_arr != 0].flatten()
    thresh_val = np.quantile(vess_vals_arr, q=min_quantile)
    vess_mask_img = input_image.threshold_image(low_thresh=thresh_val, high_thresh=None)

    vess_mask_img[:, :, :z_crop] = 0

    if morph_dil_radius > 0:
        vess_mask_img = vess_mask_img.morphology(operation='dilate', radius=morph_dil_radius)
    return vess_mask_img


def unique_segmentation_labels(segmentation_img: ants.core.ANTsImage | np.ndarray,
                               zeroth_roi: bool=False) -> np.ndarray:
    """
    Get a list of unique ROI labels in a segmentation image.

    Args:
        segmentation_img (ants.core.ANTsImage | np.ndarray): Segmentation image or array with
            integer labels for distinct ROIs.
        zeroth_roi (bool): Determines whether region index zero should be included as a unique
            region. If True, ``0`` is included in output list, otherwise is not included. Default
            False.

    Returns:
        labels (np.ndarray): Array of unique integers in the segmentation image.
    """
    if isinstance(segmentation_img, ants.core.ANTsImage):
        segmentation_arr = segmentation_img.numpy()
    else:
        segmentation_arr = segmentation_img
    labels = np.unique(segmentation_arr)
    labels = labels.astype(np.uint32)
    if not zeroth_roi:
        labels = labels[labels != 0]
    return labels

def seg_crop_to_pet_fov(pet_img: ants.ANTsImage,
                        segmentation_img: ants.ANTsImage,
                        pet_thresh_value: float=np.finfo(float).eps) -> ants.ANTsImage:
    """Zero out segmentation values that lie outside of the PET FOV.
    
    Especially applicable to scanners with limited FOV (field of view). PET voxels with values less
    than 1e-36 are considered outside of the FOV.
    
    Args:
        pet_img (ants.ANTsImage): PET image in anatomical space used to crop segmentation
        segmentation_img (ants.ANTsImage): Segmentation image in anatomical space such as
            FreeSurfer to which FOV cropping is applied.
        pet_thresh_value (float): Lower threshold for the PET image by which the segmentation image
            is masked. Should be <<1. Default machine epsilon for `float`.

    Returns:
        segmentation_masked_img (ants.ANTsImage): Segmentation image masked to PET FOV.
    """
    if not check_physical_space_for_ants_image_pair(pet_img, segmentation_img):
        raise ValueError("PET and segmentation image must share physical space.")
    pet_mean_img = get_average_of_timeseries(input_image=pet_img)
    pet_mask = ants.threshold_image(pet_mean_img, pet_thresh_value)
    seg_masked = ants.mask_image(segmentation_img, pet_mask)
    return seg_masked



"""Convert image list to timeseries image."""
import ants
import numpy as np


def gen_nd_image_based_on_image_list(image_list: list[ants.ANTsImage]) -> ants.ANTsImage:
    r"""
    Generate a 4D ANTsImage based on a list of 3D ANTsImages.

    This function takes a list of 3D ANTsImages and constructs a new 4D ANTsImage,
    where the additional dimension represents the number of frames (3D images) in the list.
    The 4D image retains the spacing, origin, direction, and shape properties of the 3D images,
    with appropriate modifications for the additional dimension.

    Args:
        image_list (list[ants.core.ants_image.ANTsImage]):
            List of 3D ANTsImage objects to be combined into a 4D image.
            The list must contain at least one image, and all images must have the same
            dimensions and properties.

    Returns:
        ants.ANTsImage:
            A 4D ANTsImage constructed from the input list of 3D images. The additional
            dimension corresponds to the number of frames (length of the image list).

    Raises:
        AssertionError: If the `image_list` is empty or if the images in the list are not 3D.

    See Also
        * :func:`petpal.preproc.motion_corr.motion_corr_frame_list_to_t1`

    Example:

        .. code-block:: python


            import ants
            image1 = ants.image_read('frame1.nii.gz')
            image2 = ants.image_read('frame2.nii.gz')
            image_list = [image1, image2]
            result = _gen_nd_image_based_on_image_list(image_list)
            print(result.dimension)  # 4
            image4d = ants.list_to_ndimage(result, image_list)

    """
    assert len(image_list) > 0
    assert image_list[0].dimension == 3

    num_frames = len(image_list)
    spacing_3d = image_list[0].spacing
    origin_3d = image_list[0].origin
    shape_3d = image_list[0].shape
    direction_3d = image_list[0].direction

    direction_4d = np.eye(4)
    direction_4d[:3, :3] = direction_3d
    spacing_4d = (*spacing_3d, 1.0)
    origin_4d = (*origin_3d, 0.0)
    shape_4d = (*shape_3d, num_frames)

    tmp_image = ants.make_image(imagesize=shape_4d,
                                spacing=spacing_4d,
                                origin=origin_4d,
                                direction=direction_4d)
    return tmp_image


def timeseries_from_img_list(image_list: list[ants.core.ANTsImage]) -> ants.core.ANTsImage:
    r"""
    Takes a list of ANTs ndimages, and generates a 4D ndimage. Undoes :func:`ants.ndimage_to_list`
    so that we take a list of 3D images and generates a 4D image.

    Args:
        image_list (list[ants.core.ANTsImage]): A list of ndimages.

    Returns:
        ants.core.ANTsImage: 4D ndimage.
    """
    tmp_image = gen_nd_image_based_on_image_list(image_list)
    return ants.list_to_ndimage(tmp_image, image_list)


def gen_3d_img_from_timeseries(input_img: ants.ANTsImage) -> ants.ANTsImage:
    """
    Get the first frame of a 4D image as a template 3D image with voxel value zero.

    A simplified version of :py:func:`ants.ndimage_to_list.ndimage_to_list`.

    Args:
        input_img (ants.ANTsImage): The 4D image from which to get the template image.

    Returns:
        img_3d (ants.ANTsImage): The 3D template of the input image as an ants image.
    """
    dimension = input_img.dimension
    subdimension = dimension - 1
    suborigin = ants.get_origin( input_img )[0:subdimension]
    subspacing = ants.get_spacing( input_img )[0:subdimension]
    subdirection = np.eye( subdimension )
    for i in range( subdimension ):
        subdirection[i,:] = ants.get_direction( input_img )[i,0:subdimension]
    img_shape = input_img.shape[:-1]
    img_3d = ants.make_image(img_shape)
    ants.set_spacing( img_3d, subspacing )
    ants.set_origin( img_3d, suborigin )
    ants.set_direction( img_3d, subdirection )

    return img_3d


def get_frame_from_timeseries(input_img: ants.ANTsImage, frame: int) -> ants.ANTsImage:
    """
    Get a single frame of a 4D image as a 3D image.

    A simplified version of :py:func:`ants.ndimage_to_list.ndimage_to_list`.

    Args:
        input_img (ants.ANTsImage): The 4D image from which to get the frame.
        frame (int): The index of the frame to extract from the time series image.

    Returns:
        img_3d (ants.ANTsImage): The 3D first frame of the input image as an ants image.
    """
    dimension = input_img.dimension
    subdimension = dimension - 1
    suborigin = ants.get_origin( input_img )[0:subdimension]
    subspacing = ants.get_spacing( input_img )[0:subdimension]
    subdirection = np.eye( subdimension )
    for i in range( subdimension ):
        subdirection[i,:] = ants.get_direction( input_img )[i,0:subdimension]
    img_3d = ants.slice_image( input_img, axis = subdimension, idx = frame )
    ants.set_spacing( img_3d, subspacing )
    ants.set_origin( img_3d, suborigin )
    ants.set_direction( img_3d, subdirection )

    return img_3d


def check_physical_space_for_ants_image_pair(image_1: ants.core.ANTsImage,
                                             image_2: ants.core.ANTsImage,
                                             tolerance: float=1e-2) -> bool:
    """
    Determines whether two ANTs images share the same physical space. This function works
    when comparing 4D-images with 3D-images, as opposed to
    :func:`ants.image_physical_space_consistency`.

    This function validates whether the direction matrices, spacing values, and origins
    of the two provided ANTs images are consistent, ensuring they reside in the same
    physical space.

    Args:
        image_1 (ants.core.ANTsImage): The first ANTs image for comparison.
        image_2 (ants.core.ANTsImage): The second ANTs image for comparison.
        tolerance (float): Absolute tolerance for differences between components of the affine
            transform matrix for the two images. Default 0.01.

    Returns:
        bool: `True` if both images share the same physical space, `False` otherwise.

    """


    dir_cons = np.allclose(image_1.direction[:3,:3], image_2.direction[:3,:3],atol=tolerance)
    spc_cons = np.allclose(image_1.spacing[:3], image_2.spacing[:3],atol=tolerance)
    org_cons = np.allclose(image_1.origin[:3], image_2.origin[:3],atol=tolerance)

    return dir_cons and spc_cons and org_cons



"""
CLI module to plot TACs from a TACs folder created by petpal function write-tacs.
"""
import argparse

from ..visualizations.tac_plots import TacFigure, RegionalTacFigure
from ..utils.time_activity_curve import TimeActivityCurve


_PLOT_EXAMPLES_ = r"""
Examples:
  - Plot a single TAC:
    petpal-plot-tacs --tac-files my_tac.tsv --out-fig-path tac.png
  - Plot two or more TACs:
    petpal-plot-tacs --tac-files my_tac_1.tsv my_tac_2.tsv --out-fig-path tac.png
  - Plot all the TACs in a directory:
    petpal-plot-tacs --tac-dir sub-001/tacs/ --out-fig-path tac.png
  - Plot specific regional TACs in a directory based on region names:
    petpal-plot-tacs --tac-files my_tac.tsv --regions RightPutamen LeftPutamen --out-fig-path tac.png
  - Set x-axis and y-axis units:
    petpal-plot-tacs --tac-files my_tac.tsv --yaxis-units cps --xaxis-units hours --out-fig-path tac.png
  - Plot the linear-linear plot only:
    petpal-plot-tacs --tac-files my_tac.tsv --plot-type linear --out-fig-path tac.png
  - Set the figure title to the participant name:
    petpal-plot-tacs --tac-files my_tac.tsv --fig-title sub-001 --out-fig-path tac.png
"""


def main():
    """
    CLI for tac plotting
    """

    parser = argparse.ArgumentParser(prog='petpal-plot-tacs',
                                     description='Command line interface for plotting TACs.',
                                     formatter_class=argparse.RawTextHelpFormatter,
                                     epilog=_PLOT_EXAMPLES_)
    parser.add_argument('--tac-files',
                        required=False,
                        nargs='+',
                        help='Path to one or more individual .tsv TAC files, separate paths with '
                             'spaces. At least one of: --tac-files, --tac-dir, is required.')
    parser.add_argument('--tac-dir',
                        required=False,
                        help='Path to a directory containing .tsv TAC files generated with PETPAL.'
                             ' At least one of: --tac-files, --tac-dir, is required.')
    parser.add_argument('--out-fig-path',
                        required=True,
                        help='Path to the file where the figure is saved.')
    parser.add_argument('--fig-title',
                        required=False,
                        help='Sets the figure title.')
    parser.add_argument('--regions',
                        required=False,
                        nargs='+',
                        help='If --tac-files is set, list the regions to be plotted. Separate '
                             'region names with spaces. Expecting TAC file names to follow the '
                             'convention: *seg-SegmentName* where SegmentName does not contain '
                             'special characters, especially - and _ which will conflict with '
                             'PETPAL code.')
    parser.add_argument('--plot-type',
                        required=False,
                        default='both',
                        choices=['linear','log','both'],
                        help='Set whether to plot the TACs as linear-linear, log-linear, or both.')
    parser.add_argument('--yaxis-units',
                        required=False,
                        default='Bq/mL',
                        choices=['Bq/mL','kBq/mL','cps','nCi/mL'],
                        help='Set activity concentration unit label for the y-axis. Does not scale'
                             ' units, this only assigns the axis label name.')
    parser.add_argument('--xaxis-units',
                        required=False,
                        default='minutes',
                        choices=['minutes','seconds','hours'],
                        help='Set time units for the x-axis. Does not scale units, this only '
                             'assigns the axis label name.')
    parser.add_argument('--plot-style',
                        required=False,
                        default='lines',
                        choices=['lines','markers','both'],
                        help='Set style of TAC plots.')
    parser.add_argument('--errorbars',
                        required=False,
                        action=argparse.BooleanOptionalAction,
                        help='Optional setting that adds errorbars to plots, if uncertainties are'
                             'available in the TAC files.')

    args = parser.parse_args()
    all_tac_plot_func = None
    regions_plot_func = None


    if args.tac_dir is None and args.tac_files is None:
        raise SystemExit('Both --tac-files and --tac-dir unset. Exiting.')

    if args.tac_dir is None:
        fig = TacFigure(plot_type=args.plot_type,
                        xlabel=fr'$t$ [{args.xaxis_units}]',
                        ylabel=fr'TAC [$\mathrm{{{args.yaxis_units}}}$]')
    else:
        fig = RegionalTacFigure(tacs_dir=args.tac_dir,
                                plot_type=args.plot_type,
                                xlabel=fr'$t$ [{args.xaxis_units}]',
                                ylabel=fr'TAC [$\mathrm{{{args.yaxis_units}}}$]')
        if args.errorbars is None:
            all_tac_plot_func = fig.plot_all_regional_tacs
            regions_plot_func = fig.plot_tacs_in_regions_list
        else:
            all_tac_plot_func = fig.plot_all_regional_tacs_with_errorbar
            regions_plot_func = fig.plot_tacs_in_regions_list_with_errorbar

    plot_style_opts = {'lines': {'marker':'None','ls':'-'},
                       'markers': {'marker':'.','ls':'None'},
                       'both': {'marker':'.','ls':'-'}}
    plot_style = plot_style_opts[args.plot_style]

    if args.tac_files is not None:
        if args.errorbars is None:
            for tac_file in args.tac_files:
                tac = TimeActivityCurve.from_tsv(filename=tac_file)
                fig.add_tac(*tac.tac, **plot_style)
        else:
            for tac_file in args.tac_files:
                tac = TimeActivityCurve.from_tsv(filename=tac_file)
                fig.add_errorbar(*tac.tac_werr, **plot_style)

    if args.tac_dir is not None:
        if args.regions is None:
            all_tac_plot_func(**plot_style)
        else:
            regions_plot_func(regions=args.regions, **plot_style)

    if args.fig_title is not None:
        fig.fig.suptitle(t=args.fig_title)

    fig.write_fig(out_fig_path=args.out_fig_path)


if __name__ == "__main__":
    main()



"""
This module contains the plotting functionalities for the project. It contains various classes that handle different
types of plots.

Classes defined in this module are:
    - :class:`GraphicalAnalysisPlot`: It handles various parameters for graphical plot analysis. Provides
      functionalities for generating figures, adding plot data, shading, and more.
    - :class:`PatlakPlot`: This class is responsible for plotting functionalities specific to Patlak method. It includes
      methods for calculating indexes, generating labels, and more.
    - :class:`LoganPlot`: This class handles plotting functionalities for the Logan method. It also includes methods for
      calculating indexes, generating labels, etc.
    - :class:`AltLoganPlot`: This class is primarily for enhancing the LoganPlot with additional or alternate features.
    - :class:`Plot`: A general plot class that handles common plot factors such as output directory, method name, ROI
      TAC paths, etc. It includes utilities for validating file path, directory and more.

Functions and methods in this module use :mod:`matplotlib` and :mod:`numpy` packages for creating and manipulating
plots.

Please refer to the documentation of each class for more detailed information.
"""

from abc import ABC, abstractmethod
import os
from typing import Dict, Union, Type
from matplotlib import pyplot as plt
import numpy as np
from ..kinetic_modeling import graphical_analysis as pet_grph
from ..utils.time_activity_curve import safe_load_tac


class GraphicalAnalysisPlot(ABC):
    """
    This is an abstract base class designed for creating customizable plots for graphical analysis.
    It takes Time-Activity Curves (TACs) as input and creates various plots (x vs. y, fit lines, fit area
    shading plots, and fit points) based on user input. This class also calculates fit parameters for the plotting data
    and determines relevant indices based on given thresholds.

    Attributes:
        pTAC (np.ndarray): The Input/Plasma TAC, an array containing time points and corresponding activity.
        tTAC (np.ndarray): The Tissue/Region TAC, an array with time points and corresponding activity.
        t_thresh_in_mins (float): The threshold time, in minutes, to consider in the analysis. Points are fit after this
            threshold.
        fig (plt.Figure): A matplotlib Figure object where the plots will be drawn.
        ax_list (list): A list of matplotlib Axes associated with `fig` where the plots will be drawn.
        non_zero_idx (np.ndarray): Indexes of non-zero values in appropriate TAC (calculated in specific
            implementations).
        t_thresh_idx (int): The index at which the time threshold is crossed in the TACs (calculated in specific
            implementations).
        x (np.ndarray): The "x" values for plotting (calculated in specific implementations).
        y (np.ndarray): The "y" values for plotting (calculated in specific implementations).
        fit_params (dict): The parameters fit to the data using least squares. Contains 'slope', 'intercept', and
            'r_squared' (calculated in specific implementations).

    .. important::

        This is an abstract class and should be inherited by a concrete class that implements the following methods:
            * :func:`calculate_valid_indicies_and_x_and_y`
            * :func:`generate_label_from_fit_params`
            * :func:`add_figure_axes_labels_and_legend`

    """

    def __init__(self, pTAC: np.ndarray, tTAC: np.ndarray, t_thresh_in_mins: float, figObj: plt.Figure = None):
        """
        Initialize an instance of the GraphicalAnalysisPlot class.

        The instance is initialized with two Time-Activity Curves (TACs), a threshold time, and an optional matplotlib
        Figure. It calculates valid indices (where the denominator is non-zero for the particular analysis), 'x' and 'y'
        values for plotting based on the TACs and the threshold time, and also analyzes the TACs to generate the fits.

        Args:
            pTAC (np.ndarray): The input Time Activity Curve, an array containing time points and corresponding activity.
            tTAC (np.ndarray): The Tissue or Region Time Activity Curve, an array with time points and corresponding
                activity.
            t_thresh_in_mins (float): The threshold time in minutes to consider when performing calculations for the
                plots.
            figObj (plt.Figure, optional): An optional matplotlib Figure object. If not provided, a new Figure object is
                created.

        Raises:
            matplotlib error: Error handling for the plot generation is managed by matplotlib. Any exceptions thrown
                during plotting are handled by the :mod:`matplotlib` internally.
        """
        self.pTAC: np.ndarray = pTAC[:]
        self.tTAC: np.ndarray = tTAC[:]
        self.t_thresh_in_mins: float = t_thresh_in_mins
        self.fig, self.ax_list = self.generate_figure_and_axes(figObj=figObj)
        self.non_zero_idx: np.ndarray = None
        self.t_thresh_idx: int = None
        self.x: np.ndarray = None
        self.y: np.ndarray = None
        self.fit_params: Dict = None
        self.calculate_valid_indicies_and_x_and_y()
        self.calculate_fit_params()

    @staticmethod
    def generate_figure_and_axes(figObj: plt.Figure = None):
        """
        Generate a matplotlib Figure and Axes for plotting.

        A new Figure and Axes are created if no Figure object is provided. If a Figure object is provided, the method
        retrieves the existing axes from the Figure object. In either case, the method returns the Figure and a list of
        its Axes.

        Args:
            figObj (plt.Figure, optional): An optional matplotlib Figure object. If not provided, a new Figure object is
                created with 2 subplots arranged in 1 row, a figure size of 8x4, line width of 3.0, and edge color 'k'.

        Returns:
            fig (plt.Figure): The resulting matplotlib Figure object.
            ax_list (list): A list of Axes objects associated with 'fig'.

        Raises:
            None
        """
        if figObj is None:
            fig, ax_list = plt.subplots(1, 2,
                                        constrained_layout=True, figsize=[8, 4],
                                        linewidth=3.0, edgecolor='k')
            ax_list = ax_list.flatten()
        else:
            fig = figObj
            ax_list = fig.get_axes()
        return fig, ax_list

    def add_data_plots(self, pl_kwargs: dict = None):
        """
        Add data plots to the Axes in the instance's Axes list.

        This method plots the instance's :math:`x` and :math:`y` values (from the particular analysis) on each Axes in
        the instance's Axes list. The style of the plotted points can be customized by passing a dictionary of keyword
        arguments for `matplotlib.pyplot.plot`.

        Args:
            pl_kwargs (dict, optional): A dictionary of keyword arguments to be passed to `matplotlib.pyplot.plot` for
                styling the points. If not provided, the points are plotted with ``linewidth=1``, ``alpha=0.9``,
                ``markersize=8``, ``markerstyle='.'``, ``zorder=1``, and ``color='black'``.

        Raises:
            ValueError: If pl_kwargs contains an argument not supported by ``matplotlib.pyplot.plot``.
        """
        if pl_kwargs is None:
            for ax in self.ax_list:
                ax.plot(self.x, self.y, lw=1, alpha=0.9, ms=8, marker='.', zorder=1, color='black')
        else:
            for ax in self.ax_list:
                ax.plot(self.x, self.y, **pl_kwargs)

    def add_shading_plots(self, pl_kwargs: dict = None):
        """
        Add shaded regions to the Axes in the instance's Axes list.

        This method shades a vertical span on each Axes in the instance's Axes list, from the x-value at the threshold
        index to the last x-value. The style of the shaded region can be customized by passing a dictionary of keyword
        arguments for :func:`matplotlib.pyplot.axvspan`.

        Args:
            pl_kwargs (dict, optional): A dictionary of keyword arguments to be passed to `matplotlib.pyplot.axvspan`
                for styling the shaded region. If not provided, the region is shaded with ``color='gray'``,
                ``alpha=0.2``, and ``zorder=0``.

        Raises:
            ValueError: If pl_kwargs contains an argument not supported by :func:`matplotlib.pyplot.axvspan`.
        """
        x_lo, x_hi = self.x[self.t_thresh_idx], self.x[-1]
        if pl_kwargs is None:
            for ax in self.ax_list:
                ax.axvspan(x_lo, x_hi, color='gray', alpha=0.2, zorder=0)
        else:
            for ax in self.ax_list:
                ax.axvspan(x_lo, x_hi, **pl_kwargs)

    def add_fit_points(self, pl_kwargs: dict = None):
        """
        Add fit points to the Axes in the instance's Axes list.

        This method plots the instance's :math:`x` and :math:`y` values that were used in fitting, i.e., the points past
        the threshold index, on each Axes in the instance's Axes list. The style of the plotted points can be customized
        by passing a dictionary of keyword arguments for :func:`matplotlib.pyplot.plot`.

        Args:
            pl_kwargs (dict, optional): A dictionary of keyword arguments to be passed to :func:`matplotlib.pyplot.plot`
                for styling the points. If not provided, the points are plotted as blue circles with ``alpha=0.9``,
                ``markersize=5``, and ``zorder=2``.

        Raises:
            ValueError: If pl_kwargs contains an argument not supported by :func:`matplotlib.pyplot.plot`.
        """
        t_thresh = self.t_thresh_idx
        if pl_kwargs is None:
            for ax in self.ax_list:
                ax.plot(self.x[t_thresh:], self.y[t_thresh:], 'o', alpha=0.9, ms='5', zorder=2, color='blue')
        else:
            for ax in self.ax_list:
                ax.plot(self.x[t_thresh:], self.y[t_thresh:], **pl_kwargs)

    def add_fit_lines(self, pl_kwargs: dict = None):
        """
        Add line of best fit to the Axes in the instance's Axes list.

        This method plots the line of best fit based on the instance's fit parameters (slope and intercept) on each
        Axes in the instance's Axes list. The style of the line can be customized by passing a dictionary of keyword
        arguments for :func:`matplotlib.pyplot.plot`.

        Args:
            pl_kwargs (dict, optional): A dictionary of keyword arguments to be passed to :func:`matplotlib.pyplot.plot`
                for styling of the line. If not provided, the line is plotted as an orange solid line with line
                ``linewidth=2.5``, ``zorder=3``, and labelled using the :meth:`generate_label_from_fit_params` method.

        Raises:
            ValueError: If pl_kwargs contains an argument not supported by :func:`matplotlib.pyplot.plot`.
        """
        y = self.x * self.fit_params['slope'] + self.fit_params['intercept']
        if pl_kwargs is None:
            for ax in self.ax_list:
                ax.plot(self.x, y, '-', color='orange', lw=2.5,
                        zorder=3, label=self.generate_label_from_fit_params())
        else:
            for ax in self.ax_list:
                ax.plot(self.x, y, **pl_kwargs)

    def add_plots(self,
                  plot_data: bool = True,
                  plot_fit_points: bool = True,
                  plot_fit_lines: bool = True,
                  fit_shading: bool = True,
                  data_kwargs: dict = None,
                  points_kwargs: dict = None,
                  line_kwargs: dict = None,
                  shading_kwargs: dict = None):
        """
        Generate and add various types of plots to the Axes in the instance's Axes list.

        Depending on the boolean values of ``plot_data``, ``plot_fit_points``, ``plot_fit_lines``, and ``fit_shading``,
        this method determines which types of plots to generate and include. This includes the data points plot,
        fitting points, line of best fit, and shading between the line of best fit and the x-axis.

        Args:
            plot_data (bool, optional): Determines if the analysis :math:`x` and :math:`y` points should be plotted.
            Defaults to True.
            plot_fit_points (bool, optional): Determines if the fit points should be plotted. Defaults to True.
            plot_fit_lines (bool, optional): Determines if the line of best fit should be plotted. Defaults to True.
            fit_shading (bool, optional): Determines if shading should be applied to the plots. Defaults to True.
            data_kwargs (dict, optional): Keyword arguments for styling the data plot.
            points_kwargs (dict, optional): Keyword arguments for styling the fit points plot.
            line_kwargs (dict, optional): Keyword arguments for styling the line of best fit.
            shading_kwargs (dict, optional): Keyword arguments for styling the shading.

        Raises:
            ValueError: If any of the keyword argument dictionaries contain unsupported arguments.

        See Also:
            * :meth:`add_data_plots`: Method used to plot the :math:`x` and :math:`y` data.
            * :meth:`add_fit_points`: Method used to plot the fit points.
            * :meth:`add_fit_lines`: Method used to plot the line of best fit.
            * :meth:`add_shading_plots`: Method used to add shading to the plots.

        """
        if plot_data:
            self.add_data_plots(pl_kwargs=data_kwargs)
        if plot_fit_points:
            self.add_fit_points(pl_kwargs=points_kwargs)
        if plot_fit_lines:
            self.add_fit_lines(pl_kwargs=line_kwargs)
        if fit_shading:
            self.add_shading_plots(pl_kwargs=shading_kwargs)

    def generate_figure(self,
                        plot_data: bool = True,
                        plot_fit_points: bool = True,
                        plot_fit_lines: bool = True,
                        fit_shading: bool = True,
                        data_kwargs: dict = None,
                        points_kwargs: dict = None,
                        line_kwargs: dict = None,
                        shading_kwargs: dict = None):
        """
        Generates the complete figure for graphical analysis.

        This function is the preferred way for a user to interact with this class. It sets up and adds the desired plots
        to the figure, adds the labels and legend, and sets the title and scale for the plots. By default, we generate
        a figure where each panel has the analysis :math:`x` and :math:`y` data, the points used for the fitting, a
        shading over the region for the fit, and the fit line.

        Args:
            plot_data (bool, optional): Determines if the :math:`x` and :math:`y` data points should be plotted.
            Defaults to True.
            plot_fit_points (bool, optional): Determines if the fit points should be plotted. Defaults to True.
            plot_fit_lines (bool, optional): Determines if the line of best fit should be plotted. Defaults to True.
            fit_shading (bool, optional): Determines if shading should be applied to the plots. Defaults to True.
            data_kwargs (dict, optional): Keyword arguments for styling the data plot.
            points_kwargs (dict, optional): Keyword arguments for styling the fit points plot.
            line_kwargs (dict, optional): Keyword arguments for styling the line of best fit.
            shading_kwargs (dict, optional): Keyword arguments for styling the shading.

        See Also:
            * :func:`add_plots`: Composite function that adds different types of plots.
            * :func:`add_figure_axes_labels_and_legend`: Adds labels and a legend to the figure.
              (To be implemented in a specific class)

        """
        self.add_plots(plot_data=plot_data,
                       plot_fit_points=plot_fit_points,
                       plot_fit_lines=plot_fit_lines,
                       fit_shading=fit_shading,
                       data_kwargs=data_kwargs,
                       points_kwargs=points_kwargs,
                       line_kwargs=line_kwargs,
                       shading_kwargs=shading_kwargs)
        self.add_figure_axes_labels_and_legend()
        self.ax_list[0].set_title("Linear Plot")
        self.ax_list[1].set_title("LogLog Plot")

        self.ax_list[1].set(yscale='log', xscale='log')

    def calculate_fit_params(self):
        """
        Calculates the parameters (slope, intercept, r_squared) for line fitting.

        This function fits a line to the :math:`x` and :math:`y` values that are generated, beyond the provided threshold.
        The fit line is computed using Least Square fitting with R-squared method.

        See Also:
            * :func:`calculate_valid_indicies_and_x_and_y`: Method used to generate the x and y values.
            * :func:`petpal.graphical_analysis.fit_line_to_data_using_lls_with_rsquared`: Method used to fit a line to
              data points using Least Square fitting with R-squared value.

        """
        t_thresh = self.t_thresh_idx
        fit_params = pet_grph.fit_line_to_data_using_lls_with_rsquared(xdata=self.x[t_thresh:], ydata=self.y[t_thresh:])

        fit_params = {
            'slope': fit_params[0],
            'intercept': fit_params[1],
            'r_squared': fit_params[2]
        }
        self.fit_params = fit_params

    @abstractmethod
    def calculate_valid_indicies_and_x_and_y(self) -> None:
        """
        Abstract method for calculating the :math:`x` and :math:`y` values given a particular analysis in a concrete
        class, valid indices where the denominator is 0, and the index corresponding to the provided threshold.

        This is a two-fold method:
        1. Calculates valid x and y values derived from the analysis where each concrete class represents a different
        method. Valid points are those where the denominator (which varies by analysis type) is non-zero.
        2. Saves the indices of the non-zero valid values and the indices that correspond to the given t_thresh_in_mins.

        .. important::
            This abstract method must be implemented in each subclass representing a different analysis method.

        Note:
            The implementation of this function in subclasses MUST calculate and assign values to ``x``, ``y``,
            ``t_thresh_idx``, and ``non_zero_idx`` attributes.

        Raises:
            NotImplementedError: This method needs to be implemented in a subclass.

        Example Implementation:
            :meth:`PatlakPlot.calculate_valid_indicies_and_x_and_y`: This class provides an example implementation of
            this method in a concrete subclass.

        """
        raise NotImplementedError("This method must be implemented in a concrete class.")

    @abstractmethod
    def generate_label_from_fit_params(self) -> str:
        """
        Abstract method to generate a label string from fitting parameters.

        This function reads the fitting parameters 'slope', 'intercept', and 'r_squared' from the class attribute
        'fit_params', and formats them into a string that can be used as a label on plots or reports.

        .. important::
            This abstract method must be implemented in each subclass representing a different analysis method.

        Raises:
            NotImplementedError: This method needs to be implemented in a subclass.

        Example Implementation:
            :meth:`PatlakPlot.generate_label_from_fit_params`: This method provides an example implementation of this
            abstract method in a concrete subclass. Similarly, :meth:`LoganPlot.generate_label_from_fit_params`.

        """
        raise NotImplementedError("This method must be implemented in a concrete class.")

    @abstractmethod
    def add_figure_axes_labels_and_legend(self):
        """
        Abstract method for adding analysis specific titles, axes labels and a legend to the figure.

        .. important::
            This abstract method must be implemented in each subclass representing a different analysis method.

        Raises:
            NotImplementedError: This method must be implemented in each subclass.

        Example Implementation:
            For an example of implementing this method, see :meth:`PatlakPlot.add_figure_axes_labels_and_legend`,
            or :meth:`LoganPlot.add_figure_axes_labels_and_legend`.

        """
        raise NotImplementedError("This method must be implemented in a concrete class.")


class PatlakPlot(GraphicalAnalysisPlot):
    r"""
    This class handles generation of Patlak plots for PET analysis.

    The :class:`PatlakPlot` class is designed to process PET data and display it in the form of Patlak plots. The class
    handles data processing including computation of valid indices and calculation of plot-specific parameters, setting
    of plot labels and legends, as well as the actual plotting of data.

    The processing steps involve computations such as handling of non-zero indices in the plasma time-activity curve
    (pTAC), calculation of x and y coordinates for the Patlak plot, and generation of plot labeling information.

    The class also provides a capability to add labels, legends and title to the figure  to present the data in a
    meaningful way.

    Note:
        The class inherits from the :class:`GraphicalAnalysisPlot` abstract base class.

    Example:
        In the proceeding examples, ``pTAC`` represent the plasma TAC (or input TAC) and ``tTAC`` represents the tissue TAC.

        For the quickest way to generate the Patlak Plot, we just instantiate the class and run the :meth:`generate_figure`
        method. This will generate 2 plots in a row. The first one having a linear-linear scale, and the second one
        having a log-log scale.

        .. code-block:: python

            from petpal.graphical_plots import PatlakPlot
            patlak_plot = PatlakPlot(tTAC=tTAC, pTAC=pTAC, t_thresh_in_mins=45.0)
            patlak_plot.generate_figure()
            plt.show() # Or use plt.savefig() to save the figure.

        If the default styling needs to be changed, we can pass keyword arguments to each of the plotting methods:

        .. code-block:: python

            from petpal.graphical_plots import PatlakPlot
            patlak_plot = PatlakPlot(tTAC=tTAC, pTAC=pTAC, t_thresh_in_mins=45.0)
            patlak_plot.generate_figure(line_kwargs=dict(lw=2, alpha=0.95, color='red', label=patlak_plot.generate_label_from_fit_params()),
                                        shading_kwargs=dict(color='palegreen', alpha=0.2),
                                        data_kwargs=dict(alpha=0.85, color='k', marker='.'))

    See Also:
            * :class:`LoganPlot`
            * :class:`AltLoganPlot`

    """

    def calculate_valid_indicies_and_x_and_y(self) -> None:
        r"""
        Calculates the valid indices along with :math:`x` and :math:`y` for Patlak plot analysis.

        This method performs the computation for the non-zero indices in the provided plasma time-activity curve (pTAC).
        It further calculates the values of :math:`x` and :math:`y` used in Patlak analysis based on these non-zero
        indices. This is done to avoid singularities caused by zero denominators. The Patlak :math:`x` and :math:`y`
        values are:

        .. math::

            y&= \frac{R(t)}{C_\mathrm{P}(t)}\\
            x&= \frac{\int_{0}^{t}C_\mathrm{P}(s)\mathrm{d}s}{C_\mathrm{P}(t)},

        where :math:`C_\mathrm{P}` is the input function and :math:`R(t)` is PET activity in the particular region of
        interest.

        The method updates the instance variables ``x``, ``y``, ``non_zero_idx``, and ``t_thresh_idx``.

        Returns:
            None
        """
        non_zero_indices = np.argwhere(self.pTAC[1] != 0.0).T[0]
        t_thresh = pet_grph.get_index_from_threshold(times_in_minutes=self.pTAC[0][non_zero_indices],
                                                     t_thresh_in_minutes=self.t_thresh_in_mins)

        x = pet_grph.cumulative_trapezoidal_integral(xdata=self.pTAC[0], ydata=self.pTAC[1])
        x = x[non_zero_indices] / self.pTAC[1][non_zero_indices]
        y = self.tTAC[1][non_zero_indices] / self.pTAC[1][non_zero_indices]

        self.x = x[:]
        self.y = y[:]
        self.non_zero_idx = non_zero_indices[:]
        self.t_thresh_idx = t_thresh
        return None

    def generate_label_from_fit_params(self) -> str:
        r"""
        Creates a label string from the fit parameters for graphical presentation.

        This method retrieves slope, intercept, and R-squared values from the instance's
        fit parameters, and then formats these values into a string that is LaTeX compatible
        for later rendering inside a plot's label. For example:

        .. math::

            K_{1}&=0.1\\
            V_{\mathrm{T}}&=0.2\\
            R^{2}&=0.95

        Returns:
            str: The created label. Each parameter is formatted as a separate line.
        """
        slope = self.fit_params['slope']
        intercept = self.fit_params['intercept']
        r_sq = self.fit_params['r_squared']

        return fr"$K_1=${slope:<5.3f}\n$V_\mathrm{{T}}=${intercept:<5.3f}\n$R^2=${r_sq:<5.3f}"

    def add_figure_axes_labels_and_legend(self):
        r"""
        Adds labels and a legend to the axes of the figure.

        This method sets the `x_label` and `y_label` for all axes on the figure. It also adds a legend to the figure,
        which is anchored to the upper left corner. Lastly, we also give a title to the figure: Patlak Plots.

        The labels are set to:

        .. math::

           \begin{align*}
           y&= \frac{R(t)}{C_\mathrm{P}(t)}\\
           x&= \frac{\int_{0}^{t}C_\mathrm{P}(s)\mathrm{d}s}{C_\mathrm{P}(t)},
           \end{align*}

        where :math:`C_\mathrm{P}` is the input function and :math:`R(t)` is PET activity in the particular region of
        interest.

        See Also:
            * :meth:`calculate_valid_indicies_and_x_and_y` for the calculation implementation.

        """
        x_label = r"$\frac{\int_{0}^{t}C_\mathrm{P}(s)\mathrm{d}s}{C_\mathrm{P}(t)}$"
        y_label = r"$\frac{R(t)}{C_\mathrm{P}(t)}$"
        for ax in self.ax_list:
            ax.set_xlabel(x_label)
            ax.set_ylabel(y_label)
        self.fig.legend(*self.ax_list[0].get_legend_handles_labels(),
                        bbox_to_anchor=(1.0, 0.8),
                        loc='upper left',
                        title='Patlak Analysis')
        self.fig.suptitle("Patlak Plots")


class LoganPlot(GraphicalAnalysisPlot):
    r"""
    This class handles generation of Logan plots for PET analysis.

    The :class:`LoganPlot` class is designed to process PET data and display it in the form of Logan plots.
    The class handles data processing including computation of valid indices and calculation of plot-specific
    parameters, setting of plot labels and legends, as well as the actual plotting of data.

    The processing steps involve computations such as handling of non-zero indices in the plasma time-activity curve
    (pTAC), calculation of x and y coordinates for the Logan plot, and generation of plot labeling information.

    The class also provides a capability to add labels, legends and title to the figure to present the data in a
    meaningful way.

    Note:
        The class inherits from the :class:`GraphicalAnalysisPlot` abstract base class.

    Example:
        In the proceeding examples, ``pTAC`` represent the plasma TAC (or input TAC) and ``tTAC`` represents the tissue
        TAC.

        For the quickest way to generate the Logan Plot, we just instantiate the class and run the :meth:`generate_figure`
        method. This will generate 2 plots in a row. The first one having a linear-linear scale, and the second one
        having a log-log scale.

        .. code-block:: python

            from petpal.graphical_plots import LoganPlot
            logan_plot = LoganPlot(tTAC=tTAC, pTAC=pTAC, t_thresh_in_mins=45.0)
            logan_plot.generate_figure()
            plt.show() # Or use plt.savefig() to save the figure.

        If the default styling needs to be changed, we can pass keyword arguments to each of the plotting methods:

        .. code-block:: python

            from petpal.graphical_plots import LoganPlot
            logan_plot = LoganPlot(tTAC=tTAC, pTAC=pTAC, t_thresh_in_mins=45.0)
            logan_plot.generate_figure(line_kwargs=dict(lw=2, alpha=0.95, color='red', label=patlak_plot.generate_label_from_fit_params()),
                                        shading_kwargs=dict(color='palegreen', alpha=0.2),
                                        data_kwargs=dict(alpha=0.85, color='k', marker='.'))

    See Also:
        * :class:`PatlakPlot`
        * :class:`AltLoganPlot`

    """

    def calculate_valid_indicies_and_x_and_y(self) -> None:
        r"""Calculates the valid indices along with :math:`x` and :math:`y` for Logan plot analysis.

        This method performs the computation for the non-zero indices in the provided region time-activity curve (tTAC).
        It further calculates the values of :math:`x` and :math:`y` used in Logan analysis based on these non-zero
        indices. This is done to avoid singularities caused by zero denominators. The Logan :math:`x` and :math:`y`
        values are:

        .. math::

            y&= \frac{\int_{0}^{t}R(s)\mathrm{d}s}{R(t)}\\
            x&= \frac{\int_{0}^{t}C_\mathrm{P}(s)\mathrm{d}s}{R(t)},

        where :math:`C_\mathrm{P}` is the input function and :math:`R(t)` is PET activity in the particular region of
        interest.

        The method updates the instance variables ``x``, ``y``, ``non_zero_idx``, and ``t_thresh_idx``.

        Returns:
            None

        """
        non_zero_indices = np.argwhere(self.tTAC[1] != 0.0).T[0]
        t_thresh = pet_grph.get_index_from_threshold(times_in_minutes=self.pTAC[0][non_zero_indices],
                                                     t_thresh_in_minutes=self.t_thresh_in_mins)

        x = pet_grph.cumulative_trapezoidal_integral(xdata=self.pTAC[0], ydata=self.pTAC[1])
        y = pet_grph.cumulative_trapezoidal_integral(xdata=self.tTAC[0], ydata=self.tTAC[1])

        x = x[non_zero_indices] / self.tTAC[1][non_zero_indices]
        y = y[non_zero_indices] / self.tTAC[1][non_zero_indices]

        self.x = x[:]
        self.y = y[:]
        self.non_zero_idx = non_zero_indices[:]
        self.t_thresh_idx = t_thresh
        return None

    def generate_label_from_fit_params(self) -> str:
        r"""
        Creates a label string from the fit parameters for graphical presentation.

        This method retrieves slope, intercept, and R-squared values from the instance's fit parameters, and then
        formats these values into a string that is LaTeX compatible for later rendering inside a plot's label.
        For example:

        .. math::

            V_\mathrm{T}&=0.1\\
            b&=0.2\\
            R^{2}&=0.95

        Returns:
            str: The created label. Each parameter is formatted as a separate line.
        """
        slope = self.fit_params['slope']
        intercept = self.fit_params['intercept']
        r_sq = self.fit_params['r_squared']

        return fr"$V_\mathrm{{T}}=${slope:<5.3f}\n$b=${intercept:<5.3f}\n$R^2=${r_sq:<5.3f}"

    def add_figure_axes_labels_and_legend(self):
        r"""
        Adds labels and a legend to the axes of the figure.

        This method sets the `x_label` and `y_label` for all axes on the figure. It also adds a legend to the figure,
        which is anchored to the upper left corner. Lastly, we also give a title to the figure: Logan Plots.

        The labels are set to:

        .. math::

            y&= \frac{\int_{0}^{t}R(s)\mathrm{d}s}{R(t)}\\
            x&= \frac{\int_{0}^{t}C_\mathrm{P}(s)\mathrm{d}s}{R(t)},

        where :math:`C_\mathrm{P}` is the input function and :math:`R(t)` is PET activity in the particular region of
        interest.

        See Also:
            * :meth:`calculate_valid_indicies_and_x_and_y` for the calculation implementation.

        """
        x_label = r"$\frac{\int_{0}^{t}C_\mathrm{P}(s)\mathrm{d}s}{R(t)}$"
        y_label = r"$\frac{\int_{0}^{t}R(s)\mathrm{d}s}{R(t)}$"
        for ax in self.ax_list:
            ax.set_xlabel(x_label)
            ax.set_ylabel(y_label)
        self.fig.legend(*self.ax_list[0].get_legend_handles_labels(), bbox_to_anchor=(1.0, 0.8), loc='upper left',
                        title='Logan Analysis')
        self.fig.suptitle("Logan Plots")


class AltLoganPlot(GraphicalAnalysisPlot):
    r"""
    This class handles generation of Alternative Logan plots (or "new plots") for PET analysis.

    The :class:`AltLoganPlot` class is designed to process PET data and display it in the form of Alt-Logan plots. The
    class handles data processing including computation of valid indices and calculation of plot-specific parameters,
    setting of plot labels and legends, as well as the actual plotting of data.

    The processing steps involve computations such as handling of non-zero indices in the plasma time-activity curve
    (pTAC), calculation of :math:`x` and :math:`y` coordinates for the Alt-Logan plot, and generation of plot labeling
    information.

    The class also provides a capability to add labels, legends and title to the figure to present the data in a
    meaningful way.

    Note:
        The class inherits from the :class:`GraphicalAnalysisPlot` abstract base class.

    Example:
        In the proceeding examples, ``pTAC`` represent the plasma TAC (or input TAC) and ``tTAC`` represents the tissue TAC.

        For the quickest way to generate the Alt-Logan Plot, we just instantiate the class and run the
        :meth:`generate_figure` method. This will generate 2 plots in a row. The first one having a linear-linear scale,
        and the second one having a log-log scale.

        .. code-block:: python

            from petpal.graphical_plots import AltLoganPlot
            alt_logan_plot = AltLoganPlot(tTAC=tTAC, pTAC=pTAC, t_thresh_in_mins=45.0)
            alt_logan_plot.generate_figure()
            plt.show() # Or use plt.savefig() to save the figure.

        If the default styling needs to be changed, we can pass keyword arguments to each of the plotting methods:

        .. code-block:: python

            from petpal.graphical_plots import AltLoganPlot
            alt_logan_plot = AltLoganPlot(tTAC=tTAC, pTAC=pTAC, t_thresh_in_mins=45.0)
            alt_logan_plot.generate_figure(line_kwargs=dict(lw=2, alpha=0.95, color='red', label=patlak_plot.generate_label_from_fit_params()),
                                        shading_kwargs=dict(color='palegreen', alpha=0.2),
                                        data_kwargs=dict(alpha=0.85, color='k', marker='.'))

    See Also:
            * :class:`PatlakPlot`
            * :class:`LoganPlot`

    """

    def calculate_valid_indicies_and_x_and_y(self) -> None:
        r"""Calculates the valid indices along with :math:`x` and :math:`y` for Alt-Logan plot analysis.

        This method performs the computation for the non-zero indices in the provided region time-activity curve (tTAC).
        It further calculates the values of :math:`x` and :math:`y` used in Logan analysis based on these non-zero
        indices. This is done to avoid singularities caused by zero denominators. The Alt-Logan :math:`x` and :math:`y`
        values are:

        .. math::

            y&= \frac{\int_{0}^{t}R(s)\mathrm{d}s}{C_\mathrm{P}(t)}\\
            x&= \frac{\int_{0}^{t}C_\mathrm{P}(s)\mathrm{d}s}{C_\mathrm{P}(t)},

        where :math:`C_\mathrm{P}` is the input function and :math:`R(t)` is PET activity in the particular region of
        interest. Compared to :class:`LoganPlot`, the denominator is the input function.

        The method updates the instance variables ``x``, ``y``, ``non_zero_idx``, and ``t_thresh_idx``.

        Returns:
            None

        """
        non_zero_indices = np.argwhere(self.pTAC[1] != 0.0).T[0]
        t_thresh = pet_grph.get_index_from_threshold(times_in_minutes=self.pTAC[0][non_zero_indices],
                                                     t_thresh_in_minutes=self.t_thresh_in_mins)

        x = pet_grph.cumulative_trapezoidal_integral(xdata=self.pTAC[0], ydata=self.pTAC[1])
        y = pet_grph.cumulative_trapezoidal_integral(xdata=self.tTAC[0], ydata=self.tTAC[1])

        x = x[non_zero_indices] / self.pTAC[1][non_zero_indices]
        y = y[non_zero_indices] / self.pTAC[1][non_zero_indices]

        self.x = x[:]
        self.y = y[:]
        self.non_zero_idx = non_zero_indices[:]
        self.t_thresh_idx = t_thresh
        return None

    def generate_label_from_fit_params(self) -> str:
        r"""
        Creates a label string from the fit parameters for graphical presentation.

        This method retrieves slope, intercept, and R-squared values from the instance's fit parameters, and then
        formats these values into a string that is LaTeX compatible for later rendering inside a plot's label.
        For example:

        .. math::

            V_\mathrm{T}&=0.1\\
            b&=0.2\\
            R^{2}&=0.95

        Returns:
            str: The created label. Each parameter is formatted as a separate line.
        """
        slope = self.fit_params['slope']
        intercept = self.fit_params['intercept']
        r_sq = self.fit_params['r_squared']

        return f"$m=${slope:<5.3f}\n$b=${intercept:<5.3f}\n$R^2=${r_sq:<5.3f}"

    def add_figure_axes_labels_and_legend(self):
        r"""
        Adds labels and a legend to the axes of the figure.

        This method sets the `x_label` and `y_label` for all axes on the figure. It also adds a legend to the figure,
        which is anchored to the upper left corner. Lastly, we also give a title to the figure: Alt-Logan Plots.

        The labels are set to:

        .. math::

            y&= \frac{\int_{0}^{t}R(s)\mathrm{d}s}{C_\mathrm{P}(t)}\\
            x&= \frac{\int_{0}^{t}C_\mathrm{P}(s)\mathrm{d}s}{C_\mathrm{P}},

        where :math:`C_\mathrm{P}` is the input function and :math:`R(t)` is PET activity in the particular region of
        interest.

        See Also:
            * :meth:`calculate_valid_indicies_and_x_and_y` for the calculation implementation.

        """
        x_label = r"$\frac{\int_{0}^{t}C_\mathrm{P}(s)\mathrm{d}s}{C_\mathrm{P}(t)}$"
        y_label = r"$\frac{\int_{0}^{t}R(s)\mathrm{d}s}{C_\mathrm{P}(t)}$"
        for ax in self.ax_list:
            ax.set_xlabel(x_label)
            ax.set_ylabel(y_label)
        self.fig.legend(*self.ax_list[0].get_legend_handles_labels(), bbox_to_anchor=(1.0, 0.8), loc='upper left',
                        title='Alt-Logan Analysis')
        self.fig.suptitle("Alt-Logan Plots")


class Plot:
    """
    A class used to generate and save graphical analysis plots of PET Time-Activity Curves (TACs).

    This class provides a simple and convenient interface to generate plots from TACs for the 'patlak', 'logan', or
    'alt-logan' methods and save them in the specified output directory. The plots are saved in both PNG and PDF formats.

    It is initialized with paths to the input TACs, the method name, the output directory, and an optional filename
    prefix. During initialization, it checks the validity of the file paths and directory, loads the TACs, and selects
    the appropriate figure class based on the given method. Later on, the `save_figure` method can be called to generate
    and save the plots.

    Attributes:
        input_tac_path (str): Path to the input TAC file.
        roi_tac_path (str): Path to the Region of Interest (ROI) TAC file.
        output_directory (str): Path to the directory where the plot output will be saved.
        method_name (str): Name of the method for generating the plot ('patlak', 'logan', 'alt-logan').
        thresh_in_mins (float): A threshold in minutes below which data points will be discarded from the plot.
        output_filename_prefix (str): An optional prefix for the output filenames.

    Example:
        An example of how to use this class to save a Patlak graphical analysis plot:


        .. code-block:: python

            import petpal.graphical_plots as pet_plt
            # ptac_path points to a plasma TAC (or input TAC)
            # ttac_path points to a ROI TAC.
            grph_plot = pet_plt.Plot(input_tac_path=ptac_path,
                                     roi_tac_path=ttac_path,
                                     threshold_in_mins=30.0,
                                     method_name='patlak',
                                     output_directory='./',
                                     output_filename_prefix="plot")
            grph_plot.save_figure()

    """

    def __init__(self,
                 input_tac_path: str,
                 roi_tac_path: str,
                 threshold_in_mins: float,
                 method_name: str,
                 output_directory: str,
                 output_filename_prefix: str = "") -> None:
        """
        Initializes the :class:`Plot` object given the paths and the method.

        During the initialization, it validates the given file paths and directory,  and selects the appropriate figure
        class based on the provided method name.

        Args:
            input_tac_path (str): Path to the input Time Activity Curve (TAC) file.
            roi_tac_path (str): Path to the Region of Interest (ROI) TAC file.
            threshold_in_mins (float): Threshold in minutes below which data points will be discarded.
            method_name (str): Name of the method to be used ('patlak', 'logan', or 'alt-logan').
            output_directory (str): Path to the directory where output files will be stored.
            output_filename_prefix (str, optional): Prefix to be appended to the output files. Defaults to an empty string.

        Raises:
            ValueError: If any of the file paths or the directory path does not exist, or if the method name is invalid.

        Note:
            This method does not return a value. It's an initializer of this class.

        Calls:
            * :meth:`_validate_filepath`
            * :meth:`_validate_directory`
            * :func:`safe_load_tac`
            * :meth:`_select_fig_class_based_on_method`

        """
        self.input_tac_path = os.path.abspath(input_tac_path)
        self.roi_tac_path = os.path.abspath(roi_tac_path)
        self.output_directory = os.path.abspath(output_directory)
        self._validate_filepath(self.input_tac_path)
        self._validate_filepath(self.roi_tac_path)
        self._validate_directory(self.output_directory)
        self.method_name = method_name
        self.thresh_in_mins = threshold_in_mins
        self.output_filename_prefix = output_filename_prefix
        self.fig_cls = self._select_fig_class_based_on_method(method_name=self.method_name)

    def save_figure(self):
        """
        Creates and saves the figure in both PNG and PDF formats.

        We first safely load the TACS. Then, this method generates the figure using the method selected during the
        initialization of the class. The figure is then saved inside the specified  output directory with a filename
        based on parameters provided at initialization.

        The saved files are in both PNG (with a resolution of 150 dpi) and PDF formats (with the PDF file being saved
        with transparency).

        Note:
            This method does not return any value. It is responsible for saving the generated figure to the specified
            output directory in PNG and PDF formats.

        Raises:
            Exception: An error occurred while loading the TACs from the input or ROI files.

        """
        p_tac = safe_load_tac(self.input_tac_path)
        t_tac = safe_load_tac(self.roi_tac_path)

        filename = f"{self.output_filename_prefix}_analysis-{self.method_name}"
        out_path = os.path.join(self.output_directory, filename)

        with plt.rc_context(rc={'pdf.fonttype': 42, 'ps.fonttype': 42, 'text.usetex': True}):
            fig_cls = self.fig_cls(pTAC=p_tac, tTAC=t_tac, t_thresh_in_mins=self.thresh_in_mins)
            fig_cls.generate_figure()
            plt.savefig(f"{out_path}.png", bbox_inches='tight', dpi=150)
            plt.savefig(f"{out_path}.pdf", bbox_inches='tight', transparent=True)
            plt.close()

    @staticmethod
    def _validate_filepath(filename: str) -> None:
        """
        Checks if the provided filename is a valid file.

        This is a private static method and should be called internally by the class while processing files.

        If the file is not valid, it raises a ValueError.

        Args:
            filename (str): The path to the file.

        Raises:
            ValueError: if the file at provided path does not exist.

        No return value.

        """
        if not os.path.isfile(filename):
            raise ValueError(f"Invalid file path: {filename}")

    @staticmethod
    def _validate_directory(directory: str) -> None:
        """
        Validates if the provided directory path is an existing directory.

        This is a private static method and should be called internally by the class while processing directories.

        If the directory does not exist, it raises a ValueError.

        Args:
            directory (str): Path to the directory.

        Raises:
            ValueError: if the provided directory path does not exist.

        No return value.

        """
        if not os.path.isdir(directory):
            raise ValueError(f"Invalid directory: {directory}.")

    @staticmethod
    def _select_fig_class_based_on_method(method_name: str) -> Union[
        Type[PatlakPlot], Type[LoganPlot], Type[AltLoganPlot]]:
        """
        Selects and returns the appropriate class based on the provided method name.

        This private static method returns the corresponding class object (`PatlakPlot`, `LoganPlot`, or `AltLoganPlot`)
        according to the provided method name ('patlak', 'logan', or 'alt-logan').

        Args:
            method_name (str): Name of the method.

        Returns:
            :class:`PatlakPlot` | :class:`LoganPlot` | :class:`AltLoganPlot`: Corresponding class.

        Raises:
            ValueError: The method name is invalid.

        See Also:
            * :class:`PatlakPlot`
            * :class:`LoganPlot`
            * :class:`AltLoganPlot`

        """
        if method_name == "patlak":
            return PatlakPlot
        elif method_name == "logan":
            return LoganPlot
        elif method_name == "alt-logan":
            return AltLoganPlot
        else:
            raise ValueError("Invalid method name. Please choose either 'patlak', 'logan', or 'alt-logan'.")



from . import bids_utils
from . import image_io
from . import math_lib
from . import testing_utils
from . import useful_functions
from . import time_activity_curve
from . import decorators
from . import data_driven_image_analyses
from . import constants
from . import scan_timing
from . import metadata
from . import stats
from . import dimension

def main():
    print("PETPAL - Utilities")


if __name__ == "__main__":
    main()
    